from math import cos, sin

import numpy as np
from scipy.optimize import newton

from pysisyphus.irc.IRC import IRC
from pysisyphus.optimizers.hessian_updates import bfgs_update
from pysisyphus.TableFormatter import TableFormatter


# [1] An improved algorithm for reaction path following
# http://aip.scitation.org/doi/pdf/10.1063/1.456010
# [2] Extension to internal coordinates (not implemented)
# https://pubs.acs.org/doi/pdf/10.1021/j100377a021


class GonzalezSchlegel(IRC):
    def __init__(
        self,
        geometry,
        max_micro_cycles=20,
        micro_step_thresh=1e-3,
        hessian_recalc=None,
        line_search=False,
        **kwargs,
    ):
        super().__init__(geometry, **kwargs)

        self.max_micro_cycles = max_micro_cycles
        self.micro_step_thresh = micro_step_thresh
        self.hessian_recalc = hessian_recalc
        self.line_search = line_search
        if self.line_search:
            print("!Line search seems faulty right now!")

        self.pivot_coords = list()
        self.eye = np.eye(self.geometry.coords.size)
        self.micro_counter = 0

        micro_header = "# |dx| |tangent|".split()
        micro_fmts = ["d", ".2E", ".3E"]
        self.micro_formatter = TableFormatter(micro_header, micro_fmts, 10)

    def perp_component(self, vec, perp_to):
        # Make unit vector
        # perp_to /= np.linalg.norm(perp_to)
        # Substract parallel component
        return vec - perp_to.dot(vec) * perp_to / perp_to.dot(perp_to)#np.linalg.norm(perp_to)**2

    def micro_step(self, counter):
        """Constrained optimization on a hypersphere."""

        # Calculate gradient at current coordinates
        gradient = self.mw_gradient
        self.log(f"\tnorm(mw_grad)={np.linalg.norm(gradient):.6f}")

        # Interpolation proposed in the paper (Eq. (12) - (15)).
        # Does not seem to help.
        if self.line_search and (counter > 0):
            pivot_coords = self.pivot_coords[-1]
            p_prev = self.prev_coords - pivot_coords  # p"
            p_cur = self.mw_coords - pivot_coords  # p'
            g_prev = self.prev_grad  # g"
            g_cur = self.gradient  # g'
            g_prev_p = self.perp_component(g_prev, p_prev)
            g_cur_p = self.perp_component(g_cur, p_cur)
            g_prev_p_norm = np.linalg.norm(g_prev_p)
            g_cur_p_norm = np.linalg.norm(g_cur_p)
            # Angle between p_prev and p_cur
            theta_prime = np.arccos(
                p_prev.dot(p_cur) / np.linalg.norm(p_prev) / np.linalg.norm(p_cur)
            )  # θ'
            theta = g_prev_p_norm * theta_prime / (g_prev_p_norm - g_cur_p_norm)  # θ
            theta_quot = theta / theta_prime
            cos_theta = cos(theta)
            sin_theta = sin(theta)
            cos_theta_prime = cos(theta_prime)
            sin_theta_prime = sin(theta_prime)
            sin_quot = sin_theta / sin_theta_prime
            # Interpolated quantities
            g_interp = g_prev * (1 - theta_quot) + g_cur * theta_quot
            p_interp = (
                p_prev * (cos_theta - sin_quot * cos_theta_prime) + p_cur * sin_quot
            )
            x_interp = pivot_coords + p_interp
            gradient = g_interp
            self.mw_coords = x_interp
            self.displacement = p_interp

        gradient_diff = gradient - self.prev_grad
        coords_diff = self.mw_coords - self.prev_coords
        # Update previous quantities.
        self.prev_coords = self.mw_coords.copy()
        self.prev_grad = gradient.copy()

        # Recalculate Hessian
        if (
            self.hessian_recalc
            # and (self.micro_counter > 0)
            and (self.micro_counter % self.hessian_recalc == 0)
        ):
            self.mw_hessian = self.geometry.mw_hessian
        # Or update Hessian
        else:
            dH, _ = bfgs_update(self.mw_hessian, coords_diff, gradient_diff)
            self.mw_hessian += dH
        eigvals, eigvecs = np.linalg.eigh(self.mw_hessian)

        constraint = (0.5 * self.step_length) ** 2
        big = np.abs(eigvals) > 1e-8
        big_eigvals = eigvals[big]
        big_eigvecs = eigvecs[:, big]
        grad_star = big_eigvecs.T.dot(gradient)
        displ_star = big_eigvecs.T.dot(self.displacement)

        def get_dx(lambda_):
            """In basis of Hessian eigenvectors."""
            return -(grad_star - lambda_ * displ_star) / (big_eigvals - lambda_)

        def on_sphere(lambda_):
            p = displ_star + get_dx(lambda_)
            return p.dot(p) - constraint

        # Initial guess for λ.
        # λ must be smaller then the smallest eigenvector
        lambda_0 = big_eigvals[0]
        lambda_0 *= 1.5 if (lambda_0 < 0) else 0.5
        self.log(f"\tSmallest eigenvalue is {big_eigvals[0]:.4f}, λ_0={lambda_0:.4f}.")
        # Find the root with scipy
        lambda_ = newton(on_sphere, lambda_0, maxiter=500)
        self.log(f"\tDetermined λ={lambda_0:.4f} from Newtons method.")

        # Calculate dx from optimized lambda in basis of Hessian eigenvectors and
        # transform back to mass-weighted Cartesians.
        dx = big_eigvecs.dot(get_dx(lambda_))
        self.displacement += dx
        self.mw_coords += dx

        grad_tangent_to_sphere = self.perp_component(gradient, self.displacement)
        self.micro_counter += 1

        dx_norm = np.linalg.norm(dx)
        grad_norm = np.linalg.norm(grad_tangent_to_sphere)
        self.log(f"\tnorm(dx)={dx_norm:.6f}")
        self.log(f"\tgradient tangent to sphere={grad_norm:.6f}")

        return dx, grad_tangent_to_sphere

    def step(self):
        grad0 = self.mw_gradient
        grad0_norm = np.linalg.norm(grad0)
        # For the BFGS update in the first micro step we use the original
        # point and the initial guess to calculate gradient and
        # coordinate differences.
        self.prev_grad = grad0
        self.prev_coords = self.mw_coords

        # Take a step against the gradient to the pivot point x*_k+1.
        pivot_step = 0.5 * self.step_length * -grad0 / grad0_norm
        pivot_coords = self.mw_coords + pivot_step
        self.pivot_coords.append(pivot_coords)

        # Initial guess for x'_k+1 (full step from prev_coords, or another
        # half step from the pivot point)
        self.mw_coords = pivot_coords + pivot_step

        # Initial displacement p' from the pivot point
        self.displacement = pivot_step

        micro_coords_ = list()
        for i in range(self.max_micro_cycles):
            self.log(f"Micro cycle {i:02d}")
            try:
                dx, _ = self.micro_step(i)
            except RuntimeError:
                print("Constrained search did not converge!")
                self.converged = True
                return
            micro_coords_.append(self.mw_coords)
            if np.linalg.norm(dx) <= self.micro_step_thresh:
                break
        else:
            self.logger.warning("Max micro cycles exceeded!")


    def postprocess(self):
        self.pivot_coords = np.array(self.pivot_coords)



from typing import List, Optional

import h5py
import numpy as np

from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import log
from pysisyphus.intcoords.augment_bonds import augment_bonds
from pysisyphus.intcoords.PrimTypes import normalize_prim_input, normalize_prim_inputs
from pysisyphus.optimizers import poly_fit
from pysisyphus.optimizers.guess_hessians import ts_hessian, HessInit
from pysisyphus.optimizers.HessianOptimizer import HessianOptimizer, HessUpdate
from pysisyphus.optimizers.Optimizer import get_data_model, get_h5_group

from pysisyphus.helpers import array2string
import torch

class TSHessianOptimizer(HessianOptimizer):
    """Optimizer to find first-order saddle points."""

    valid_updates = ("bofill", "ts_bfgs", "ts_bfgs_org", "ts_bfgs_rev")

    def __init__(
        self,
        geometry: Geometry,
        roots: Optional[List[int]] = None,
        root: int = 0,
        hessian_ref: Optional[str] = None,
        rx_modes=None,
        prim_coord=None,
        rx_coords=None,
        hessian_init: HessInit = "calc",
        hessian_update: HessUpdate = "bofill",
        hessian_recalc_reset: bool = True,
        max_micro_cycles: int = 50,
        trust_radius: float = 0.3,
        trust_max: float = 0.5,
        augment_bonds: bool = False,
        min_line_search: bool = False,
        max_line_search: bool = False,
        assert_neg_eigval: bool = False,
        **kwargs,
    ) -> None:
        """Baseclass for transition state optimizers utilizing Hessian information.

        Several arguments expect a typed primitive or an iterable of typed primitives.
        A typed primitive is specified as (PrimType, int, int, ...), e.g., for a bond
        between atoms 0 and 1: (BOND, 0, 1) or for a bend between the atom triple 0, 1, 2
        as (BEND, 0, 1, 2).

        Parameters
        ----------
        geometry
            Geometry to be optimized.
        roots
            Indices of modes to maximize along, e.g., to optimize saddle points of 2nd order.
            Overrides 'root'.
        root
            Index of imaginary mode to maximize along. Shortcut for 'roots' with only one root.
        hessian_ref
            Filename pointing to a pysisyphus HDF5 Hessian.
        rx_modes : iterable of (iterable of (typed_prim, phase_factor))
            Select initial root(s) by overlap with a modes constructed from the given
            typed primitives with respective phase factors.
        prim_coord : typed_prim
            Select initial root/imaginary mode by overlap with this internal coordinate.
            Shortcut for 'rx_modes' with only one internal coordinate.
        rx_coords : iterable of (typed_prim)
            Construct imaginary mode comprising the given typed prims by modifying
            a model Hessian.
        hessian_init
            Type of initial model Hessian.
        hessian_update
            Type of Hessian update. Defaults to BFGS for minimizations and Bofill
            for saddle point searches.
        hessian_recalc_reset
            Whether to skip Hessian recalculation after reset. Undocumented.
        max_micro_cycles
            Maximum number of RS iterations.
        trust_radius
            Initial trust radius in whatever unit the optimization is carried out.
        trust_max
            Maximum trust radius.
        augment_bonds
            Try to derive additional streching coordinates from the imaginary mode.
        min_line_search
            Carry out line search along the imaginary mode.
        max_line_search
            Carry out line search in the subspace that is minimized.
        assert_neg_eigval
            Check for the existences for at least one significant negative eigenvalue.
            If enabled and no negative eigenvalue is present the optimization will be
            aborted.

        Other Parameters
        ----------------
        **kwargs
            Keyword arguments passed to the HessianOptimizer/Optimizer baseclass.
        """

        assert (
            hessian_update in self.valid_updates
        ), f"Invalid Hessian update. Please chose from: {self.valid_updates}!"

        super().__init__(
            geometry,
            hessian_init=hessian_init,
            hessian_update=hessian_update,
            trust_radius=trust_radius,
            trust_max=trust_max,
            hessian_recalc_reset=hessian_recalc_reset,
            **kwargs,
        )

        if (root is not None) and (roots is None):
            roots = [
                root,
            ]
        elif roots is None:
            roots = list()
        self.roots = roots
        self.log(f"{self.roots=}")
        self.hessian_ref = hessian_ref
        try:
            with h5py.File(self.hessian_ref, "r") as handle:
                self.hessian_ref = handle["hessian"][:]
            _ = self.geometry.coords.size
            expected_shape = (_, _)
            shape = self.hessian_ref.shape
            # Hessian is not yet converted to the correct coordinate system if
            # coord_type != cart.
            assert (
                self.geometry.coord_type == "cart"
            ), "hessian_ref with internal coordinates are not yet supported."
            assert shape == expected_shape, (
                f"Shape of reference Hessian {shape} doesn't match the expected "
                f"shape {expected_shape} of the Hessian for the current coordinates!"
            )
        except OSError as err:
            self.log(
                f"Tried to load reference Hessian from '{self.hessian_ref}' "
                "but the file could not be found."
            )
            self.hessian_ref = None
        except (ValueError, TypeError) as err:
            self.log(f"No reference Hessian provided.")

        # Select initial root according to highest contribution of 'prim_coord'
        if prim_coord is not None:
            self.log("'prim_coord' given. Overwriting/setting 'rx_modes'.")
            rx_modes = [[[prim_coord, 1.0]]]
        self.prim_coord = prim_coord

        # Construct initial imaginary mode from the given inputs while respecting
        # the given weight/phase factors.
        self.rx_modes = rx_modes

        # Construct initial imaginary mode according the primitive internals in
        # 'rx_coords' by modifying a model Hessian.
        if rx_coords is not None:
            rx_coords = normalize_prim_inputs(rx_coords)
        self.rx_coords = rx_coords

        # Bond augmentation is only useful with calculated hessians
        self.augment_bonds = augment_bonds and (self.hessian_init == "calc")
        self.min_line_search = min_line_search
        self.max_line_search = max_line_search
        self.assert_neg_eigval = assert_neg_eigval

        self.ts_modes = list()
        self.max_micro_cycles = max_micro_cycles
        self.prim_contrib_thresh = 0.05

        self.alpha0 = 1

    @property
    def root(self):
        return self.roots[0]

    @root.setter
    def root(self, root):
        raise Exception("Setting 'self.root' is deprecated. Set 'self.roots' instead.")

    @property
    def roots(self):
        return self._roots

    @roots.setter
    def roots(self, roots):
        self._roots = np.array(roots, dtype=int)

    def prepare_opt(self, *args, **kwargs):
        if self.augment_bonds:
            self.geometry = augment_bonds(self.geometry, root=self.root)
            # Update data model and HD5 shapes, as the number of coordinates
            # may have changed.
            if self.dump:
                self.data_model = get_data_model(
                    self.geometry, self.is_cos, self.max_cycles
                )
                # self.h5_group = get_h5_group(
                #     self.h5_fn, self.h5_group_name, self.data_model
                # )

        # Calculate/set initial hessian
        super().prepare_opt(*args, **kwargs)

        # Assume a guess hessian when not calculated. This hessian has to be
        # modified according to the assumed reaction coordinates.
        if self.hessian_init != "calc" and (self.rx_coords is not None):
            assert self.geometry.coord_type != "cart", (
                "Using a modified guess Hessian for TS-optimizations is "
                "only supported in redundand internal coordinates "
                "(coord_type=redund)"
            )
            prim_inds = [
                self.geometry.internal.get_index_of_typed_prim(rxc)
                for rxc in self.rx_coords
            ]
            missing_prim_inds = [
                self.rx_coords[i] for i, _ in enumerate(prim_inds) if _ is None
            ]
            assert len(missing_prim_inds) == 0, (
                "Some of the requested reaction coordinates are not defined: "
                f"{missing_prim_inds}"
            )
            self.H = ts_hessian(self.H, coord_inds=prim_inds)

        # Determiniation of initial mode either by using a provided
        # reference hessian, or by using a supplied root.

        if isinstance(self.H, torch.Tensor):
            eigvals, eigvecs = torch.linalg.eigh(self.H)
            eigvals = eigvals.cpu().numpy()
        else:
            eigvals, eigvecs = np.linalg.eigh(self.H)
        neg_inds = eigvals < -self.small_eigval_thresh
        self.log_negative_eigenvalues(eigvals, "Initial ")

        self.log("Determining initial TS mode to follow uphill.")
        # Select an initial TS-mode by highest overlap with eigenvectors from
        # reference Hessian.
        if self.hessian_ref is not None:
            eigvals_ref, eigvecs_ref = np.linalg.eigh(self.hessian_ref)
            self.log_negative_eigenvalues(eigvals_ref, "Reference ")
            assert eigvals_ref[0] < -self.small_eigval_thresh
            ref_mode = eigvecs_ref[:, 0]
            overlaps = np.einsum("ij,j->i", eigvecs.T, ref_mode)
            ovlp_str = array2string(overlaps, precision=4)
            self.log(
                "Overlaps between eigenvectors of current Hessian "
                f"TS mode from reference Hessian:"
            )
            self.log(f"\t{ovlp_str}")

            root = np.abs(overlaps).argmax()
            self.roots = [
                root,
            ]
            print(
                "Highest overlap between reference TS mode and "
                f"eigenvector {self.root:02d}."
            )
            used_str = "overlap with reference TS mode"
        # Construct an approximate initial mode according to user input
        # and calculate overlaps with the current eigenvectors.
        elif self.rx_modes is not None:
            self.log(f"Constructing reference mode, according to user input")
            assert self.geometry.coord_type != "cart"
            int_ = self.geometry.internal
            modes = list()
            for i, rx_mode in enumerate(self.rx_modes):
                mode = np.zeros_like(self.geometry.coords)
                for typed_prim, val in rx_mode:
                    typed_prim = normalize_prim_input(typed_prim)[0]
                    ind = int_.get_index_of_typed_prim(typed_prim)
                    mode[ind] = val
                    self.log(
                        f"\tIndex {ind} (coordinate {typed_prim}) set to {val:.4f}"
                    )
                mode /= np.linalg.norm(mode)
                modes.append(mode)
                mode_str = array2string(mode, precision=2)
                self.log(f"Normalized reference mode {i:02d}: {mode_str}")

            # Calculate overlaps in non-redundant subspace by zeroing overlaps
            # in the redundant subspace.
            small_inds = np.abs(eigvals) < self.small_eigval_thresh
            # Take absolute value, because sign of eigenvectors is ambiguous.
            ovlps = np.abs(np.einsum("ij,ki->kj", eigvecs, modes))
            ovlps[:, small_inds] = 0.0
            self.roots = ovlps.argmax(axis=1)
            used_str = "overlap with user-generated mode"
        else:
            used_str = f"root(s)={self.roots}"
        self.log(f"Used {used_str} to select inital TS mode.")

        # Below, some code is found, that checks if the chosen root(s) are a
        # sensible choice, i.e., if they are negative. Currently, it is commented out,
        # as we can also start from the convex region of the PES.
        #
        # Check if the selected mode (root) is a sensible choice.
        #
        # small_eigval_thresh is positive and we dont take the absolute value
        # of the eigenvalues. So we multiply small_eigval_thresh to get a
        # negative number.
        # assert (eigvals[self.roots] < -self.small_eigval_thresh).all(), (
        # "Expected negative eigenvalue(s)! Eigenvalues of selected TS-modes "
        # f"are above the the threshold of self.small_eigval_thresh:.6e}!"
        # )

        # Select an initial TS-mode by root index. self.root may have been
        # modified by using a reference hessian.
        self.ts_modes = eigvecs[:, self.roots].T
        self.ts_mode_eigvals = eigvals[self.roots]
        self.log(
            f"Using root(s) {self.roots} with eigenvalues "
            f"{array2string(self.ts_mode_eigvals, precision=6)} as TS mode.\n"
        )

    def update_ts_mode(self, eigvals, eigvecs):
        neg_eigval_inds = eigvals < -self.small_eigval_thresh
        neg_num = neg_eigval_inds.sum()
        self.log_negative_eigenvalues(eigvals)

        # When we left the convex region of the PES we only compare to other
        # imaginary modes ... is this a bad idea? Maybe we should use all modes
        # for the overlaps?!
        if (self.ts_mode_eigvals < 0).all() and neg_num > 0:
            infix = "imaginary "
            ovlp_eigvecs = eigvecs[:, :neg_num]
            eigvals = eigvals[:neg_num]
        # When the eigenvalue corresponding to the TS mode has been negative once,
        # we should not lose all negative eigenvalues. If this happens something went
        # wrong and we crash :)
        elif self.assert_neg_eigval and neg_num == 0:
            raise AssertionError(
                "Need at least 1 negative eigenvalue for TS optimization."
            )
        # Use all eigenvectors for overlaps when the eigenvalue corresponding to the TS
        # mode is still positive.
        else:
            infix = ""
            ovlp_eigvecs = eigvecs

        if (
            self.using_active_dofs
            and self.ts_modes is not None
            and self.ts_modes.shape[1] != ovlp_eigvecs.shape[0]
        ):
            self.log("Projecting previous TS mode(s) onto active DOFs.")
            if isinstance(self.ts_modes, torch.Tensor):
                self.ts_modes = torch.stack(
                    [self.active_from_full(mode) for mode in self.ts_modes]
                )
            else:
                self.ts_modes = np.stack(
                    [self.active_from_full(mode) for mode in self.ts_modes]
                )

        # Select new TS mode according to biggest overlap with previous TS mode.
        self.log(f"Overlaps of previous TS mode with current {infix}mode(s):")
        if isinstance(eigvecs, torch.Tensor):
            ovlps = torch.abs(torch.einsum("ij,ki->kj", ovlp_eigvecs, self.ts_modes))
            ovlps = ovlps.cpu().numpy()
        else:
            ovlps = np.abs(np.einsum("ij,ki->kj", ovlp_eigvecs, self.ts_modes))
        for i, ovlp in enumerate(ovlps):
            self.log(f"\tTS mode {i:02d}: {array2string(ovlp, precision=3)}")
        max_ovlp_inds = np.argmax(ovlps, axis=1)
        for i, _ in enumerate(self.ts_modes):
            max_ovlp_ind = max_ovlp_inds[i].argmax()
            self.log(
                f"Mode {i}: highest overlap: {ovlps[i, max_ovlp_ind]:.6f} with mode "
                f"{max_ovlp_ind}"
            )
        self.roots = max_ovlp_inds
        self.ts_modes = ovlp_eigvecs.T[self.roots]
        self.ts_mode_eigvals = eigvals[self.roots]

    @staticmethod
    def do_line_search(e0, e1, g0, g1, prev_step, maximize, logger=None):
        poly_fit_kwargs = {
            "e0": e0,
            "e1": e1,
            "g0": g0,
            "g1": g1,
            "maximize": maximize,
        }
        if not maximize:
            poly_fit_kwargs.update(
                {
                    "g0": prev_step.dot(g0),
                    "g1": prev_step.dot(g1),
                }
            )
        prefix = "Max" if maximize else "Min"

        fit_result = poly_fit.quartic_fit(**poly_fit_kwargs)
        fit_energy = None
        fit_grad = None
        fit_step = None
        if fit_result and (0.0 < fit_result.x <= 2.0):
            x = fit_result.x
            log(logger, f"{prefix}-subpsace interpolation succeeded: x={x:.6f}")
            fit_energy = fit_result.y
            fit_step = (1 - x) * -prev_step
            fit_grad = (1 - x) * g0 + x * g1
        return fit_energy, fit_grad, fit_step

    def step_and_grad_from_line_search(
        self,
        energy,
        gradient_trans,
        eigvecs,
        min_indices,
        max_indices,
    ):
        ip_step = np.zeros_like(gradient_trans)
        ip_gradient_trans = gradient_trans.copy()

        if self.max_line_search and self.cur_cycle > 0:
            prev_energy = self.energies[-2]
            prev_gradient = -self.forces[-2]
            try:
                prev_gradient_trans = eigvecs.T.dot(prev_gradient)
                prev_step = self.steps[-1]
                prev_step_trans = eigvecs.T.dot(prev_step)
            # Will be raised when coordinates were rebuilt and the array shapes differe.
            except ValueError:
                return ip_step, ip_gradient_trans

            # Max subspace
            # max_energy, max_gradient, max_step = self.do_max_line_search(
            max_energy, max_gradient, max_step = self.do_line_search(
                prev_energy,
                energy,
                prev_gradient_trans[max_indices],
                gradient_trans[max_indices],
                prev_step=prev_step_trans[max_indices],
                maximize=True,
                logger=self.logger,
            )
            if max_gradient is not None:
                ip_gradient_trans[max_indices] = max_gradient
                ip_step[max_indices] = max_step

        if self.min_line_search and self.cur_cycle > 0:
            # Min subspace
            # min_energy, min_gradient, min_step = self.do_min_line_search(
            min_energy, min_gradient, min_step = self.do_line_search(
                prev_energy,
                energy,
                prev_gradient_trans[min_indices],
                gradient_trans[min_indices],
                prev_step=prev_step_trans[min_indices],
                maximize=False,
                logger=self.logger,
            )
            if min_gradient is not None:
                ip_gradient_trans[min_indices] = min_gradient
                ip_step[min_indices] = min_step
        return ip_step, ip_gradient_trans



from math import pi
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike

from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.constants import KB, AU2J
from pysisyphus.Geometry import Geometry
from pysisyphus.intcoords.PrimTypes import prims_from_prim_inputs
from pysisyphus.intcoords.update import correct_dihedrals
from pysisyphus.intcoords import Torsion
from pysisyphus.linalg import rmsd_grad
from pysisyphus.calculators.DFTD3 import DFTD3

class LogFermi:
    def __init__(self, beta, radius, T=300, origin=(0.0, 0.0, 0.0), geom=None):
        """As described in the XTB docs.

        https://xtb-docs.readthedocs.io/en/latest/xcontrol.html#confining-in-a-cavity
        """
        self.beta = beta
        self.radius = radius
        self.T = T
        self.origin = np.array(origin)

        # In Hartree
        self.kT = KB * self.T / AU2J

    def calc(self, coords3d, gradient=False):
        t0 = coords3d - self.origin[None, :]
        t1 = np.linalg.norm(t0, axis=1)
        t2 = np.exp(self.beta * (t1 - self.radius))

        energy = (self.kT * np.log(1 + t2)).sum()
        if not gradient:
            return energy

        grad = self.kT * ((self.beta * t2) / ((1 + t2) * t1))[:, None] * t0
        return energy, grad.flatten()

    def __repr__(self):
        return (
            f"LogFermi(beta={self.beta:.6f}, radius={self.radius:.6f}, "
            f"T={self.T:.6f}, origin={self.origin})"
        )


class HarmonicSphere:
    def __init__(self, k, radius, origin=(0.0, 0.0, 0.0), geom=None):
        self.k = k
        self.radius = radius
        self.origin = np.array(origin)

    def calc(self, coords3d, gradient=False):
        c3d_wrt_origin = coords3d - self.origin
        distances = np.linalg.norm(c3d_wrt_origin, axis=1)
        energies = np.where(distances > self.radius, self.k * distances**2, 0.0)
        energy = energies.sum()

        if not gradient:
            return energy

        """
        E(r(x)) = k*r**2
        dE(r(x))/dx = dE/dr * dr/dx
        dE/dr = 2*k*r
        dr/dx = x/r
        dE/dr * dr/dx = 2*k*x
        """
        grad = np.where(distances > self.radius, 2 * self.k * c3d_wrt_origin, 0.0)

        return energy, grad.flatten()

    @property
    def surface_area(self):
        """In Bohr**2"""
        return 4 * pi * self.radius**2

    def instant_pressure(self, coords3d):
        _, gradient = self.calc(coords3d, gradient=True)
        norm = np.linalg.norm(gradient)
        p = norm / self.surface_area
        return p


class Restraint:
    def __init__(self, restraints, geom=None):
        self.restraints = list()

        for prim_inp, *rest in restraints:
            prims = prims_from_prim_inputs((prim_inp,))
            assert len(prims) == 1
            prim = prims[0]
            force_const = rest.pop(0)

            try:
                ref_val = rest.pop(0)
            except IndexError:
                assert (
                    geom is not None
                ), "Need initial coordinates when no reference value is specified!"
                ref_val = prim.calculate(geom.coords3d)

            self.restraints.append((prim, force_const, ref_val))

    @staticmethod
    def calc_prim_restraint(prim, coords3d, force_const, ref_val):
        val, grad = prim.calculate(coords3d, gradient=True)
        if isinstance(prim, Torsion):
            # correct_dihedrals always returns a 1d array, even for scalar inputs
            val = correct_dihedrals(val, ref_val)[0]
        diff = val - ref_val
        pot = force_const * diff**2
        pot_grad = 2 * force_const * diff * grad
        return pot, pot_grad

    def calc(self, coords3d, gradient=False):
        energy = 0.0
        grad = np.zeros(coords3d.size)

        for prim, force_const, ref_val in self.restraints:
            penergy, pgrad = self.calc_prim_restraint(
                prim, coords3d, force_const, ref_val
            )
            energy += penergy
            grad += pgrad

        if not gradient:
            return energy

        return energy, grad.flatten()


class RMSD:
    def __init__(
        self,
        geom: Geometry,
        k: float,
        beta: float = 0.5,
        atom_indices: Optional[ArrayLike] = None,
    ):
        """Restrain based on RMSD with a reference geometry.

        As described in https://doi.org/10.1021/acs.jctc.0c01306, Eq. (5).

        Parameters
        ----------

        geom:
            Reference geometry for RMSD calculation.
        k:
            Gaussian height in units of energy. Should be a negative number if the system
            under study should stay close to the reference geometry (pulling k).
            A positive Gaussian height k results in forces that push the system under study
            away from the reference geometry (pushing k).
        b:
            Gaussian width in inverse units of lengths.
        atom_indices
            Optional, numpy array or iterable of integer atom indices. Restricts the RMSD
            calculation to these atoms. If omitted, all atoms are used.
        """

        self.k = k
        self.beta = beta
        natoms = len(geom.atoms)
        if atom_indices is None:
            atom_indices = np.arange(natoms)
        self.atom_indices = np.array(atom_indices, dtype=int)
        assert (self.atom_indices.min() >= 0) and (
            self.atom_indices.max() <= (natoms - 1)
        ), "Got atom_indices outside the valid range!"
        self.ref_coords3d = geom.coords3d.copy()
        # print("Using {self.atom_indices.size} of {natoms} atoms for RMSD calculations.")

    def calc(self, coords3d, gradient=False):
        rmsd, grad_ = rmsd_grad(
            coords3d[self.atom_indices], self.ref_coords3d[self.atom_indices]
        )

        k = self.k
        beta = self.beta
        energy = k * np.exp(-beta * rmsd**2)

        if not gradient:
            return energy

        grad = np.zeros_like(self.ref_coords3d)
        grad[self.atom_indices] = -beta * 2 * grad_ * energy
        return energy, grad.flatten()

    def __repr__(self):
        return f"RMSD(k={self.k:.4f}, beta={self.beta:.6f}, {self.atom_indices.size} atoms)"


class ExternalPotential(Calculator):

    available_potentials = {
        "logfermi": LogFermi,
        "harmonic_sphere": HarmonicSphere,
        "restraint": Restraint,
        "rmsd": RMSD,
        "d3": DFTD3,
    }

    def __init__(self, calculator=None, potentials=None, geom=None, **kwargs):
        super().__init__(**kwargs)

        self.calculator = calculator

        self.potentials = list()
        self.log("Creating external potentials")
        for i, pot_kwargs in enumerate(potentials):
            pot_kwargs.update({"geom": geom})
            pot_key = pot_kwargs.pop("type")
            pot_cls = self.available_potentials[pot_key]
            pot = pot_cls(**pot_kwargs)
            self.potentials.append(pot)
            self.log(f"\t{i:02d}: {pot}")

    def get_potential_energy(self, coords):
        coords3d = coords.reshape(-1, 3)
        potential_energies = [pot.calc(coords3d) for pot in self.potentials]
        potential_energy = sum(potential_energies)
        self.log(f"Energies from external potential: {potential_energies}")
        return potential_energy

    def get_energy(self, atoms, coords):
        potential_energy = self.get_potential_energy(coords)
        if self.calculator is not None:
            results = self.calculator.get_energy(atoms, coords)
        else:
            results = {"energy": 0.0}
        results["energy"] += potential_energy
        return results

    def get_potential_forces(self, coords):
        coords3d = coords.reshape(-1, 3)
        energies_gradients = [
            pot.calc(coords3d, gradient=True) for pot in self.potentials
        ]
        energies, gradients = zip(*energies_gradients)
        self.log(f"Energies from external potential: {energies}")
        energy = sum(energies)
        forces = -np.sum(gradients, axis=0)
        self.log(f"Forces from external potential: {forces}")
        return energy, forces

    def get_forces(self, atoms, coords):
        potential_energy, potential_forces = self.get_potential_forces(coords)
        if self.calculator is not None:
            results = self.calculator.get_forces(atoms, coords)
        else:
            results = {"energy": 0.0, "forces": np.zeros_like(coords)}
        results["energy"] += potential_energy
        results["forces"] += potential_forces
        return results

    def get_hessian(self, atoms, coords):
        raise Exception("Hessian is not implemented for ExternalPotential!")

    def __str__(self):
        pots = ", ".join([str(pot) for pot in self.potentials])
        return f"ExternalPotential({pots})"



import numpy as np
import sympy as sym

# from sympy import sympify, lambdify,

from pysisyphus.calculators.Calculator import Calculator

from pysisyphus.calculators import ORCA, Turbomole, DFTD4


CALC_CLASSES = {
    "dftd4": DFTD4.DFTD4,
    "orca": ORCA.ORCA,
    "turbomole": Turbomole.Turbomole,
}
try:
    from pysisyphus.calculators import PySCF

    CALC_CLASSES["pyscf"] = PySCF.PySCF
except (ModuleNotFoundError, OSError):
    pass


class Composite(Calculator):
    def __init__(
        self, final, keys_calcs=None, calcs=None, remove_translation=False, **kwargs
    ):
        # Either directly supply a dictionary with Calculator objects (key_calcs)
        # or a dictionary containing information to set up calculators.
        assert keys_calcs or calcs

        super().__init__(**kwargs)

        if calcs:
            keys_calcs = {}
            calc_kwargs_ = {
                "charge": self.charge,
                "mult": self.mult,
                "pal": self.pal,
                "mem": self.mem,
            }
            for i, (key, kwargs) in enumerate(calcs.items()):
                calc_kwargs = calc_kwargs_.copy()
                calc_kwargs["calc_number"] = i
                calc_kwargs["base_name"] = f"composite_{key}"
                # Don't modify original dict
                kwargs = kwargs.copy()
                type_ = kwargs.pop("type")
                calc_kwargs.update(**kwargs)
                calc_cls = CALC_CLASSES[type_]
                keys_calcs[key] = calc_cls(**calc_kwargs)

        self.keys_calcs = keys_calcs
        assert all([key in final for key in self.keys_calcs.keys()])
        self.final = final
        self.remove_translation = remove_translation

        # The energies are just numbers that we can easily substitute in
        self.energy_expr = sym.sympify(self.final)
        # The forces/Hessians are matrices that we can't just easily substitute in.
        self.arr_args = sym.symbols(" ".join(self.keys_calcs.keys()))
        self.arr_expr = sym.lambdify(self.arr_args, self.energy_expr)

    def get_final_energy(self, energies):
        return float(self.energy_expr.subs(energies).evalf())

    def get_energy(self, atoms, coords, **prepare_kwargs):
        energies = {}
        for key, calc in self.keys_calcs.items():
            energy = calc.get_energy(atoms, coords, **prepare_kwargs)["energy"]
            energies[key] = energy

        final_energy = self.get_final_energy(energies)
        results = {
            "energy": final_energy,
        }
        return results

    def get_forces(self, atoms, coords, **prepare_kwargs):
        energies = {}
        forces = {}
        for key, calc in self.keys_calcs.items():
            results = calc.get_forces(atoms, coords, **prepare_kwargs)
            energies[key] = results["energy"]
            forces[key] = results["forces"]
        keys = self.keys_calcs.keys()
        for key in keys:
            self.log(f"|forces_{key}|={np.linalg.norm(forces[key]):.6f}")
        self.log("")

        final_energy = self.get_final_energy(energies)
        final_forces = self.arr_expr(**forces)

        # Remove overall translation
        if self.remove_translation:
            f3d = final_forces.reshape(-1, 3)
            f3d -= f3d.mean(axis=0)[None, :]

        results = {
            "energy": final_energy,
            "forces": final_forces,
        }
        return results

    def get_hessian(self, atoms, coords, **prepare_kwargs):
        energies = {}
        hessians = {}
        for key, calc in self.keys_calcs.items():
            results = calc.get_hessian(atoms, coords, **prepare_kwargs)
            energies[key] = results["energy"]
            hessians[key] = results["hessian"]

        final_energy = self.get_final_energy(energies)
        final_hessian = self.arr_expr(**hessians)

        results = {
            "energy": final_energy,
            "hessian": final_hessian,
        }
        return results

    def run_calculation(self, atoms, coords, **prepare_kwargs):
        return self.get_energy(atoms, coords, **prepare_kwargs)



# tests/test_summary_log.py
"""Tests for pdb2reaction.summary_log formatting helpers."""

from pathlib import Path

import pytest

pytest.importorskip("fairchem", reason="fairchem-core not available (GPU-only)")

from pdb2reaction.summary_log import (  # noqa: E402
    _fmt_bool,
    _shorten_path,
    _format_energy_rows,
    _format_bond_changes,
)


class TestFmtBool:
    def test_true(self):
        assert _fmt_bool(True) == "True"

    def test_false(self):
        assert _fmt_bool(False) == "False"

    def test_none(self):
        assert _fmt_bool(None) == "-"

    def test_truthy_int(self):
        assert _fmt_bool(1) == "True"

    def test_falsy_zero(self):
        assert _fmt_bool(0) == "False"


class TestShortenPath:
    def test_none_path(self):
        assert _shorten_path(None, None) == "(not available)"

    def test_empty_path(self):
        # Path("") is truthy in Python (resolves to "."), so _shorten_path returns "."
        result = _shorten_path(Path(""), None)
        assert isinstance(result, str)

    def test_relative_to_root(self):
        result = _shorten_path(Path("/a/b/c/file.txt"), Path("/a/b/c"))
        assert result == "file.txt"

    def test_relative_to_parent(self):
        result = _shorten_path(Path("/a/b/c/file.txt"), Path("/a/b/c/sub"))
        assert result == "file.txt"

    def test_no_root(self):
        p = Path("/some/absolute/path.txt")
        assert _shorten_path(p, None) == str(p)


class TestFormatEnergyRows:
    def test_basic_formatting(self):
        rows = _format_energy_rows(
            labels=["R", "TS", "P"],
            energies_au=[-100.0, -99.5, -100.1],
            energies_kcal=[0.0, 313.75, -62.75],
        )
        assert len(rows) == 3
        assert "R" in rows[0]
        assert "TS" in rows[1]
        assert "P" in rows[2]

    def test_none_energies(self):
        rows = _format_energy_rows(
            labels=["R"],
            energies_au=None,
            energies_kcal=None,
        )
        assert len(rows) == 1
        assert "n/a" in rows[0]

    def test_auto_relative_energy(self):
        """When energies_kcal is None, relative is computed from AU."""
        rows = _format_energy_rows(
            labels=["R", "TS"],
            energies_au=[-100.0, -99.9],
            energies_kcal=None,
        )
        assert len(rows) == 2
        # R should have 0.0 relative (base - base = 0)
        # TS should have a positive relative energy


class TestFormatBondChanges:
    def test_empty_text(self):
        lines = _format_bond_changes("")
        assert len(lines) == 1
        assert "no covalent changes detected" in lines[0]

    def test_nonempty_text(self):
        lines = _format_bond_changes("C1-O2 formed\nN3-H4 broken")
        assert len(lines) == 2
        assert "C1-O2 formed" in lines[0]
        assert "N3-H4 broken" in lines[1]



import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.linalg import norm3


class BondedFragment(Primitive):
    def __init__(self, indices, bond_indices, **kwargs):
        from_frag, to_ = bond_indices
        assert from_frag in indices
        assert to_ not in indices
        self.bond_indices = list(bond_indices)
        kwargs["calc_kwargs"] = ("bond_indices",)
        super().__init__(indices, **kwargs)

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, bond_indices=None):
        from_frag, to_ = bond_indices
        bond = coords3d[to_] - coords3d[from_frag]
        value = norm3(bond)

        if gradient:
            bond_normed = bond / value
            row = np.zeros_like(coords3d)
            row[indices, :] = -bond_normed
            row = row.flatten()
            return value, row
        return value



# [1] https://doi.org/10.1063/1.3463717
#     Comparisons of classical and Wigner sampling of transition state
#     energy levels for quasiclassical trajectory chemical dynamics simulations
#     Sun, Hase, 2010
# [2] https://doi.org/10.1002/qua.25049
#     Effects of different initial condition samplings on photodynamics and
#     spectrum of pyrrole
#     Barbatti, Sen, 2015
# [3] https://doi.org/10.1063/1.453761
#     The Morse oscillator in position space, momentum space, and phase space
#     Dahl, Springborg, 1988
#
# An interesting paper also seems to be (but i never read it ...)
#
# [4] https://doi.org/10.1063/5.0039592
#     Sampling initial positions and momenta for nuclear trajectories
#     from quantum mechanical distributions
#     Yao, Hase, Granucci, Persico
# [5] https://doi.org/10.1039/C8CP03273D
#     Finite-temperature Wigner phase-space sampling and temperature effects on the
#     excited-state dynamics of 2-nitronaphthalene.
#     Zobel, Nogueira, Gonzalez

import argparse
import functools
from math import exp, pi
from typing import Callable, Optional
import sys

import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import NDArray
from numpy.polynomial.laguerre import Laguerre

from pysisyphus.constants import AMU2AU, AU2EV, AU2SEC, C, KB, PLANCK
from pysisyphus.helpers_pure import eigval_to_wavenumber
from pysisyphus.io import geom_from_hessian
from pysisyphus.Geometry import Geometry, get_trans_rot_projector


# From cm⁻¹ to angular frequency in atomic units
# cm⁻¹ * 100 -> m⁻¹
# m⁻¹ * C -> s⁻¹
NU2ANGFREQAU = 2 * pi * 100 * C * AU2SEC


def get_vib_state(wavenumber: float, temperature: Optional[float] = None) -> int:
    """Return random vibrational state n for given wavenumber and temperature."""
    if temperature is None:
        return 0  # Ground state

    freq = wavenumber * 100 * C  # from cm⁻¹ to s⁻¹
    vib_en_J = PLANCK * freq  # Energy in J
    quot = vib_en_J / (KB * temperature)
    Z = np.exp(-quot / 2) / (1 - np.exp(-quot))  # Partition function
    _1overZ = 1 / Z
    # TODO: check for sensible values?

    def get_p(n):
        """Probability of vibrational state with quantum number n.

        Given by a Boltzmann distribution:

                   exp(-e_n / kT)
        p_n = ----------------------------
              sum_{n = 0}^N exp(-e_n / kT)

        for a harmonic oscillator e_n = h * freq * (n + 1/2)

        See also https://chemistry.stackexchange.com/a/61120.

        """
        return np.exp(-quot * (n + 0.5)) * _1overZ

    # Determine sensible maximum vibrational state n.
    n = 0
    probabilities = list()
    probability_sum = 0.0
    # 1.0 may never be reached, so we stop earlier.
    thresh = 0.999999
    while True:
        p = get_p(n)
        probabilities.append(p)
        probability_sum += p
        if probability_sum >= thresh:
            break
        n += 1

    # Generate random number that is smaller than the current sum.
    while True:
        # Sample from the possible interval
        random_state = np.random.random() * thresh
        if random_state < probability_sum:
            break

    cumsum = np.cumsum(probabilities)
    for n, cs in enumerate(cumsum):
        if cs >= random_state:
            break
    return n


def normal_mode_reduced_masses(masses_rep, normal_modes):
    return 1 / (np.square(normal_modes) / masses_rep[:, None]).sum(axis=0)


@functools.singledispatch
def get_wigner_sampler(
    coords3d: NDArray[float],
    masses: NDArray[float],
    hessian: NDArray[float],
    temperature: Optional[float] = None,
    nu_thresh: float = 20.0,
    stddevs: float = 6.0,
) -> Callable:
    assert coords3d.shape == (len(masses), 3)
    assert hessian.shape == (coords3d.size, coords3d.size)

    # Projector to remove translation & rotation
    Proj = get_trans_rot_projector(coords3d, masses, full=True)
    masses_rep = np.repeat(masses, 3)
    PM = Proj @ np.diag(1 / np.sqrt(masses_rep))

    # Diagonalize projected, mass-weighted Hessian.
    w, v = np.linalg.eigh(PM @ hessian @ PM.T)
    nus = eigval_to_wavenumber(w)

    # TODO: dectect if system is linear and don't hardcode dropping of first 6 modes.
    # small_nu_mask = np.abs(nus) < nu_thresh
    small_nu_mask = np.zeros_like(nus, dtype=bool)
    small_nu_mask[:6] = True
    # w = w[~small_nu_mask]
    v = v[:, ~small_nu_mask]
    nus = nus[~small_nu_mask]
    # Square root of angular frequencies in atomic units. Required to convert the
    # dimensionless Q and P values into atomic units.
    ang_freqs_au_sqrt = np.sqrt(nus * NU2ANGFREQAU)

    assert (nus >= nu_thresh).all(), "Imaginary wavenumbers are not yet handled!"
    nnus = len(nus)  # Number of non-zero wavenumbers

    span = 2 * stddevs
    # Pre-calculate some of the Laguerre polynomials
    # We use some shortcuts for n == 0 and n == 1
    laguerres = {
        0: lambda _: 1.0,
        1: lambda x: 1.0 - x,
    }

    def get_laguerre(n):
        lag_coeffs = np.zeros(n + 1)
        lag_coeffs[n] = 1.0
        return Laguerre(lag_coeffs)

    for i in range(2, 11):
        laguerres[i] = get_laguerre(i)

    mm_sqrt_au = np.sqrt(masses_rep * AMU2AU)
    M_inv_au = np.diag(1 / mm_sqrt_au)

    def sampler():
        Qs = np.zeros(nnus)
        Ps = np.zeros(nnus)
        for i in range(nnus):
            n = get_vib_state(nus[i], temperature)
            try:
                lag = laguerres[n]
            except KeyError:
                lag = get_laguerre(n)
                laguerres[n] = lag
            # According to eq. (31) in [3], the absolute value of the Wigner function is
            # bound between 0 and 1/π (see also Figure 3 in [3]). By omitting 1/π in the
            # Wigner function (eq. (28) in [3]) its absolute value will be between 0 and 1.
            # This saves us some multiplications/division as with 1/π it would also be sensible
            # to map 'ref' from [0, 1) to [0, 1/π)].
            prefact = (-1) ** n
            while True:
                q, p, ref = np.random.random_sample(3)
                # Map q and p from [0, 1) onto chosen interval [-stddevs, stddevs)
                q = q * span - stddevs
                p = p * span - stddevs
                r2 = q**2 + p**2
                # Wigner function. Eq. (2) in [2] or (28) in [3]. Both equations
                # differ in the presence or absence of a n! term. Here and in [3]
                # the n! is absorbed into the Laguerre polynomial.
                p_wig = prefact * exp(-r2) * lag(2 * r2)
                # We don't use p and q that correspond to a negative value of the
                # Wigner function. See the SI of [5] for a discussion.
                if 0.0 < p_wig >= ref:
                    break
            Qs[i] = q
            Ps[i] = p

        # The actual displacements/momenta depend on the vibrational frequencies.
        # Now we the dimensionless units to atomic units. See eq. (5) in [1].
        Qs /= ang_freqs_au_sqrt
        Ps *= ang_freqs_au_sqrt

        # Convert to Cartesian coordinates
        displ = M_inv_au @ v @ Qs
        velocities = M_inv_au @ v @ Ps

        # The COM remains unaffected, as we displace along vibrations,
        # not translations.
        displ_coords3d = coords3d.copy() + displ.reshape(-1, 3)

        # Remove rotation & translation from velocities at new coordinates.
        P = get_trans_rot_projector(displ_coords3d, masses, full=True)
        velocities = P.dot(velocities)
        return displ_coords3d, velocities.reshape(-1, 3)

    return sampler


@get_wigner_sampler.register
def _(geom: Geometry, **kwargs):
    return get_wigner_sampler(geom.coords3d, geom.masses, geom.hessian, **kwargs)


@get_wigner_sampler.register
def _(h5_fn: str, **kwargs):
    geom = geom_from_hessian(h5_fn)
    return get_wigner_sampler(geom)


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("h5_fn", type=str, help="Filename of pysisyphus HDF5 Hessian.")
    parser.add_argument("-n", type=int, default=100, help="Number of samples.")
    parser.add_argument("-T", type=float, default=None, help="Temperature in K.")
    parser.add_argument("--plotekin", action="store_true", help="Plot kinetic energy.")
    return parser.parse_args(args)


def run():
    args = parse_args(sys.argv[1:])

    h5_fn = args.h5_fn
    n = args.n
    temperature = args.T
    plotekin = args.plotekin

    geom = geom_from_hessian(h5_fn)

    sampler = get_wigner_sampler(geom, temperature=temperature)
    xyzs = list()
    velocities = np.zeros((n, len(geom.atoms), 3))
    for i in range(n):
        c3d, velocities[i] = sampler()
        xyzs.append(geom.as_xyz(cart_coords=c3d))

    trj_fn = f"samples_{n}_trj.xyz"
    with open(trj_fn, "w") as handle:
        handle.write("\n".join(xyzs))

    if plotekin:
        half_masses_au = geom.masses * AMU2AU / 2

        def E_kin(v_au):
            return (half_masses_au[:, None] * (v_au**2)).sum()

        E_kins = np.array([E_kin(v) for v in velocities])
        E_kins_eV = E_kins * AU2EV
        _, ax = plt.subplots()
        ax.hist(E_kins_eV, bins=100)
        plt.show()



import numpy as np

from pysisyphus.optimizers.Optimizer import Optimizer


class FIRE(Optimizer):
    # https://doi.org/10.1103/PhysRevLett.97.170201

    def __init__(
        self,
        geometry,
        dt=0.1,
        dt_max=1,
        N_acc=2,
        f_inc=1.1,
        f_acc=0.99,
        f_dec=0.5,
        n_reset=0,
        a_start=0.1,
        **kwargs,
    ):
        self.dt = dt
        self.dt_max = dt_max
        # Accelerate after N_acc cycles
        self.N_acc = N_acc
        self.f_inc = f_inc
        self.f_acc = f_acc
        self.f_dec = f_dec
        self.n_reset = n_reset
        self.a_start = a_start

        self.a = self.a_start
        # The current velocity
        self.v = np.zeros_like(geometry.coords)
        # Store the velocities for every step
        self.velocities = [
            self.v,
        ]
        self.time_deltas = [
            self.dt,
        ]

        super().__init__(geometry, **kwargs)

    def _get_opt_restart_info(self):
        opt_restart_info = {
            "a": self.a,
            "dt": self.dt,
            "n_reset": self.n_reset,
            "time_delta": self.time_deltas[-1],
            "velocity": self.velocities[-1].tolist(),
        }
        return opt_restart_info

    def _set_opt_restart_info(self, opt_restart_info):
        self.a = opt_restart_info["a"]
        self.dt = opt_restart_info["dt"]
        self.n_reset = opt_restart_info["n_reset"]
        self.time_deltas = [
            opt_restart_info["time_delta"],
        ]
        velocity = np.array(opt_restart_info["velocity"], dtype=float)
        self.velocities = [
            velocity,
        ]

    def reset(self):
        pass

    def optimize(self):
        if self.is_cos and self.align:
            (self.v,), _, _ = self.fit_rigid(vectors=(self.v,))

        self.forces.append(self.geometry.forces)
        self.energies.append(self.geometry.energy)
        forces = self.forces[-1]
        mixed_v = (
            # As 'a' gets bigger we keep less old v.
            (1.0 - self.a) * self.v
            +
            # As 'a' gets bigger we use more new v
            # from the current forces.
            self.a * np.sqrt(np.dot(self.v, self.v) / np.dot(forces, forces)) * forces
        )
        last_v = self.velocities[-1]
        # Check if forces are still aligned with the last velocity
        if (self.cur_cycle > 0) and (np.dot(last_v, forces) > 0):
            if self.n_reset > self.N_acc:
                self.dt = min(self.dt * self.f_inc, self.dt_max)
                self.a *= self.f_acc
            self.n_reset += 1
        else:
            # Reset everything when 'forces' and 'last_v' aren't
            # aligned anymore.
            mixed_v = np.zeros_like(forces)
            self.log("resetted velocities")
            self.a = self.a_start
            self.dt *= self.f_acc
            self.n_reset = 0

        v = mixed_v + self.dt * forces
        self.velocities.append(v)
        self.time_deltas.append(self.dt)
        steps = self.dt * v
        steps = self.scale_by_max_step(steps)

        velo_norm = np.linalg.norm(v)
        self.log(f"dt = {self.dt:.4f}, norm(v) {velo_norm:.4f}")

        return steps

    def __str__(self):
        return "FIRE optimizer"



import numpy as np

from pysisyphus.constants import ANG2BOHR, AU2KJPERMOL
from pysisyphus.calculators.Calculator import Calculator
from pysisyphus.calculators.LennardJones import LennardJones

# [1] https://aip.scitation.org/doi/abs/10.1063/1.445869
#     Jorgensen, 1983
# [2] https://pubs.acs.org/doi/pdf/10.1021/ja00344a001
#     Jorgensen, 1983
# Not yet implemented
#   OPC3, https://www.ncbi.nlm.nih.gov/pmc/articles/PMC4991989/


class TIP3P(Calculator):
    """Transferable Intermolecular Potential 3 Point"""
    rOH = 0.9572 * ANG2BOHR
    aHOH = 104.52  # deg
    # Charges
    qH = +0.4170
    qO = -2*qH
    sigma = 3.15061 * ANG2BOHR
    epsilon = 0.6364 / AU2KJPERMOL

    # rc = 5 Å in Bohr
    def __init__(self, rc=9.44863062728914):
        super().__init__()

        # Cutoff distance
        self.rc = rc
        self.lj = LennardJones(sigma=self.sigma, epsilon=self.epsilon, rc=self.rc)

        self.charges = np.array((self.qO, self.qH, self.qH))
        """
        coulomb_energy = (multiple of elem. charge * multiple of elem. charge)
                         / (distance in bohr) * 1 / (4 * pi * vacuum permittivity)

        coulomb_prefactor converts everything to atomic units and it is ... drum
        roll ... 1.
        from scipy.constants import value as pcval
        self.coulomb_prefactor = (1 / (4 * np.pi) * pcval("elementary charge")**2
                                  / pcval("Hartree energy") / pcval("Bohr radius")
                                  / pcval("vacuum electric permittivity")
        )
        """
        self.coulomb_prefactor = 1

    def coulomb(self, coords3d):
        waters = coords3d.shape[0] // 3
        if waters == 1:
            return 0., np.zeros((3, 3))
        stencil = np.array((0, 1, 2))
        # Pair indices of the interacting water molecules
        a, b = np.triu_indices(waters, 1)

        # Pair indices of the respective interacting atoms
        # a: 0, 1, 2, 0, 1, 2, 0, 1, 2, ...
        a = (3*np.repeat(a, 3)[:, None] + stencil).flatten()
        # b: 3, 3, 3, 4, 4, 4, 5, 5, 5, ... 
        b = np.repeat((3 * b[:, None] + stencil).flatten(), 3)

        # Pair coordinate differences
        diffs = coords3d[a] - coords3d[b]  # Shape: (Interacting pairs, 3)
        # Distances
        rs = np.linalg.norm(diffs, axis=1)

        # Assemble water charges for all atoms/coordinates
        charges = np.tile(self.charges, coords3d.shape[0] // 3)
        # Keep the uncontracted pair-energies, as we still need them for the
        # gradient.
        pair_energies = self.coulomb_prefactor * charges[a] * charges[b] / rs
        energy = pair_energies.sum()

        # See the LennardJones calculator for the explicit derivation of the
        # derivative of 1/r**n.
        products = (pair_energies / rs**2)[:, None] * diffs

        gradient = np.zeros_like(coords3d)
        # Every pair (a, b) contributes to the total gradient of atoms a and b.
        for i, prod in enumerate(products):
            gradient[a[i]] -= prod
            gradient[b[i]] += prod

        return energy, -gradient

    def calculate(self, coords3d):
        coulomb_energy, coulomb_forces = self.coulomb(coords3d)

        # Lennard-Jones interaction between Oxygens only
        o_coords3d = coords3d[::3]
        lj_energy, lj_forces = self.lj.calculate(o_coords3d)

        forces = np.zeros_like(coords3d)
        forces[::3] += lj_forces

        forces = coulomb_forces.copy()
        # Add LennardJones forces to oxygens
        forces[::3] += lj_forces
        energy = coulomb_energy + lj_energy

        return energy, forces

    def get_energy(self, atoms, coords):
        energy, _ = self.calculate(coords.reshape(-1, 3))
        return {"energy": energy}

    def get_forces(self, atoms, coords):
        energy, forces = self.calculate(coords.reshape(-1, 3))
        return {"energy": energy,
                "forces": forces.flatten(),
        }



# [1] https://doi.org/10.1002/ange.201914943
#     Heavy-Atom Tunneling in Organic Reactions
#     Castro, Karney, 2020
# [2] https://doi.org/10.1002/wcms.1165
#     Theory and simulation of atom tunneling in chemical reactions
#     Kästner, 2014
# [3] https://doi.org/10.1002/qua.25686
#     Eyringpy
#     Dzib, Merino et al, 2018
# [4] https://pubs.acs.org/doi/pdf/10.1021/j100809a040
#     TUNNELLING CORRECTIONS FOR UNSYMMETRICAL ECKART POTENTIAL ENERGY BARRIERS
#     Johnston, Heicklen, 1961
# [5] https://dx.doi.org/10.6028%2Fjres.086.014
#     A Method of Calculating Tunneling Corrections For Eckart Potential Barriers
#     Brown, 1981

from dataclasses import dataclass
from math import sin
from typing import Optional

import jinja2
import numpy as np
import numpy.typing as npt
import scipy.integrate as integrate

from pysisyphus.constants import AU2KJPERMOL, AU2SEC, C, KB, KBAU, PLANCK, PLANCKAU
from pysisyphus.io import geom_from_hessian


@dataclass
class ReactionRates:
    from_: str
    barrier: float  # in E_h
    barrier_si: float  # in kJ mol⁻¹
    temperature: float  # in K
    imag_wavenumber: float  # in cm⁻¹
    imag_frequency: float  # in s⁻¹
    rate_eyring: float  # in s⁻¹
    kappa_eyring: float
    rate_wigner: Optional[float] = None  # in s⁻¹
    kappa_wigner: Optional[float] = None
    rate_bell: Optional[float] = None  # in s⁻¹
    kappa_bell: Optional[float] = None
    rate_eckart: Optional[float] = None  # in s⁻¹
    kappa_eckart: Optional[float] = None


RX_RATES_TPL = """
             From: {{ "{: >18s}".format(rxr.from_) }} to TS
          Barrier: {{ "{: >18.6f}".format(rxr.barrier) }} E_h ( {{ rxr.barrier_si|f2 }} kJ mol⁻¹)
      Temperature: {{ "{: >18.2f}".format(rxr.temperature) }} K
Transition vector:
       Wavenumber: {{ "{: >18.2f}".format(rxr.imag_wavenumber) }} cm⁻¹
        Frequency: {{ "{: >18.6e}".format(rxr.imag_frequency) }} s⁻¹

 Type        kappa        rate / s⁻¹     rate / h⁻¹       comment
 -------  ------------ -------------- -------------- ------------------
 {{ rate_line("Eyring", rxr.kappa_eyring, rxr.rate_eyring) }}
{%- if rxr.rate_wigner %}
 {{ rate_line("Wigner", rxr.kappa_wigner, rxr.rate_wigner, "1d tunnel corr.") }}
{%- endif %}
{%- if rxr.rate_bell %}
 {{ rate_line("Bell", rxr.kappa_bell, rxr.rate_bell, "1d tunnel corr.") }}
{%- endif %}
{%- if rxr.rate_eckart %}
 {{ rate_line("Eckart", rxr.kappa_eckart, rxr.rate_eckart, "1d tunnel corr.") }}
{%- endif %}
 -------  ------------ -------------- -------------- ------------------
"""


def render_rx_rates(rx_rates: ReactionRates) -> str:
    def rate_line(name, kappa, rate, comment=""):
        rate_h = rate * 3600
        return f"{name: <12s} {kappa: >8.4f} {rate: >12.8e} {rate_h: >12.8e} {comment: >18s}"

    env = jinja2.Environment()
    env.filters["f2"] = lambda _: f"{_:.2f}"
    env.filters["f4"] = lambda _: f"{_:.4f}"
    env.filters["f6"] = lambda _: f"{_:.6f}"
    env.filters["e8"] = lambda _: f"{_: >16.8e}"

    tpl = env.from_string(RX_RATES_TPL)

    rendered = tpl.render(rxr=rx_rates, rate_line=rate_line)
    return rendered


def eyring_rate(
    barrier_height: float,
    temperature: npt.ArrayLike,
    trans_coeff: float = 1.0,
) -> npt.NDArray[np.float64]:
    """Rate constant in 1/s from the Eyring equation.

    See https://pubs.acs.org/doi/10.1021/acs.organomet.8b00456
    "Reaction Rates and Barriers" on p. 3234 and eq. (8).

    Parameters
    ----------
    barrier_height
        Barrier height (energy, enthalpy, gibbs energy, ...) in Hartree.
    temperature
        Temperature in Kelvin.
    trans_coeff
        Unitless transmission coefficient, e.g., obtained from Wigner or Eckart
        correction.

    Returns
    -------
    rate
        Eyring reaction rate in 1/s.
    """
    temperature = np.array(temperature, dtype=float)
    prefactor = trans_coeff * KB * temperature / PLANCK
    rate = prefactor * np.exp(-barrier_height / (KBAU * temperature))
    return rate


def harmonic_tst_rate(
    barrier_height: float,
    temperature: float,
    rs_part_func: float,
    ts_part_func: float,
    trans_coeff: float = 1.0,
) -> float:
    """Rate constant in 1/s from harmonic TST.

    See http://dx.doi.org/10.18419/opus-9841, chapter 5. Contrary to
    the Eyring rate this function does only takes a scalar temperature as the
    partition functions are also functions of the temperature and would
    have to be recalculated for different temperatures.

    A possible extension would be to also support multiple rs/ts partition functions,
    one for each temperature.

    Parameters
    ----------
    barrier_height
        Barrier height (energy, enthalpy, gibbs energy, ...) in Hartree.
    rs_part_func
        Partition function of the reactant state.
    ts_part_func
        Partition function of the transition state.
    temperature
        Temperature in Kelvin.
    trans_coeff
        Unitless transmission coefficient, e.g., obtained from Wigner or Eckart
        correction.

    Returns
    -------
    rate
        HTST reaction rate in 1/s.
    """
    rate_eyring = eyring_rate(barrier_height, temperature, trans_coeff)
    rate = ts_part_func / rs_part_func * rate_eyring
    return rate


def wigner_corr(
    temperature: float,
    imag_frequency: float,
) -> float:
    """Tunneling correction according to Wigner.

    See https://doi.org/10.1002/qua.25686 eq. (12) and
    https://doi.org/10.1007/978-3-642-59033-7_9 for the original publication.

    Parameters
    ----------
    temperature
        Temperature in Kelvin.
    imag_frequency
        Imaginary frequency in 1/s.

    Returns
    -------
    kappa
        Unitless tunneling correction according to Wigner.
    """
    kappa = 1 + 1 / 24 * (PLANCK * abs(imag_frequency) / (KB * temperature)) ** 2
    return kappa


def bell_corr(
    temperature: float,
    imag_frequency: float,
) -> float:
    """Tunneling correction according to Bell.

    See https://onlinelibrary.wiley.com/doi/10.1002/anie.201708489 eq. (1) and
    eq. (2).

    Parameters
    ----------
    temperature
        Temperature in Kelvin.
    imag_frequency
        Imaginary frequency in 1/s.

    Returns
    -------
    kappa
        Unitless tunneling correction according to Bell. Negative kappas are
        meaningless.
    """
    u = PLANCK * abs(imag_frequency) / (KB * temperature)
    kappa = (u / 2) / sin(u / 2)
    return kappa


def eckart_corr(
    fw_barrier_height: float,
    bw_barrier_height: float,
    temperature: float,
    imag_frequency: float,
) -> float:
    """Tunneling correction according to Eckart.

    See [3], [4] and [5]. The correction should be independent of the order
    fw_barrier_height/bw_barrier_height.

    Parameters
    ----------
    fw_barrier_height
        Barrier height in forward direction in Hartree.
    bw_barrier_height
        Barrier height in backward direction in Hartree.
    temperature
        Temperature in Kelvin.
    imag_frequency
        Frequency in 1/s of the imaginary mode at the TS.

    Returns
    -------
    kappa
        Unitless tunneling correction according to Eckart.
    """
    kBT = KBAU * temperature  # Hartree
    two_pi = 2 * np.pi

    imag_frequency = abs(imag_frequency) * AU2SEC  # Convert from 1/s to 1/au
    hnu = PLANCKAU * imag_frequency
    u = hnu / kBT  # unitless
    v1 = fw_barrier_height / kBT
    v2 = bw_barrier_height / kBT
    alpha1, alpha2 = [
        two_pi * barrier / hnu for barrier in (fw_barrier_height, bw_barrier_height)
    ]
    quot = 1 / (1 / np.sqrt(alpha1) + 1 / np.sqrt(alpha2))
    d = 1 / two_pi * np.sqrt(np.abs(4 * alpha1 * alpha2 - np.pi ** 2))
    cosh_d = np.cosh(two_pi * np.abs(d))

    def eps(E):
        return (E - fw_barrier_height) / kBT

    def ai(E, vi):
        epsilon = eps(E)
        return np.sqrt(2 * (epsilon + vi) / (np.pi * u)) * quot

    def _a1(E):
        return ai(E, vi=v1)

    def _a2(E):
        return ai(E, vi=v2)

    def eckart_probability(E):
        a1 = _a1(E)
        a2 = _a2(E)
        plus = np.cosh(two_pi * (a1 + a2))
        minus = np.cosh(two_pi * (a1 - a2))
        return (plus - minus) / (plus + cosh_d)

    def eckart_func(E):
        epsilon = eps(E)
        P = eckart_probability(E)
        return P * np.exp(-epsilon)

    # Integration limits
    E_low = max(0, fw_barrier_height - bw_barrier_height)
    E_max = 1  # 1 Hartree max
    kappa, _ = integrate.quad(eckart_func, E_low, E_max, limit=250)
    # Make kappa unitless again by dividing by kBT, as we integrated over the energy.
    kappa /= kBT
    return kappa


def tunl(alph1: float, alph2: float, U: float, strict: bool = False):
    """Eckart correction factor for rate constants according to Brown.

    Python adaption of the TUNL subroutine in 4. Appendix of [5].

    Parameters
    ----------
    alph1
        Unitless barrier height descriptor. 2π V1 / (h nu*); see (2) in [5].
    alph2
        Unitless barrier heigth descriptor. 2π V2 / (h nu*); see (2) in [5].
    u
        Unitless curvature descriptor. h nu* / kT; see (2) in [5].
    strict
        If enabled, arguments are bound checked. Will raise AssertionError if
        they are out of bonds. TUNL was found to yield accurate results when
        the arguments are within bounds.

    Returns
    -------
    G
        Unitless tunneling correction according to Eckart.
    """
    if strict:
        alphs = np.array((alph1, alph2))
        assert (0.5 <= alphs).all() and (alphs <= 20).all() and 0 < U <= 16
        if (alphs >= 8).all():
            assert U <= 12
        if (alphs >= 16).all():
            assert U <= 10

    # Quadrature arguments
    x = np.array((-0.9324695, -0.6612094, -0.2386192, 0.2386192, 0.6612094, 0.9324695))
    w = np.array((0.1713245, 0.3607616, 0.4679139, 0.4679139, 0.3607616, 0.1713245))

    pi = np.pi
    upi2 = U / (2 * pi)

    C = 0.125 * pi * U * (1 / np.sqrt(alph1) + 1.0 / np.sqrt(alph2)) ** 2
    v1 = upi2 * alph1
    v2 = upi2 * alph2
    D = 4 * alph1 * alph2 - pi ** 2
    DF = np.cosh(np.sqrt(D)) if D >= 0 else np.cos(np.sqrt(-D))

    ez = -v1 if (v2 >= v1) else -v2
    eb = min(
        C * (np.log(2.0 * (1.0 + DF) / 0.014) / (2 * pi)) ** 2 - (v1 + v2) / 2, 3.2
    )
    em = (eb - ez) / 2
    ep = (eb + ez) / 2

    # Original loop
    E = em * x + ep
    A1 = pi * np.sqrt((E + v1) / C)
    A2 = pi * np.sqrt((E + v2) / C)
    fm = np.cosh(A1 - A2)
    fp = np.cosh(A1 + A2)
    G = (w * np.exp(-E) * (fp - fm) / (fp + DF)).sum()
    # Outside the loop
    G = em * G + np.exp(-eb)
    return G


def eckart_corr_brown(
    fw_barrier_height: float,
    bw_barrier_height: float,
    temperature: float,
    imag_frequency: float,
) -> float:
    """Tunneling correction according to Eckart.

    Wrapper for the TUNL subroutine as given in the appendix of [5].

    Parameters
    ----------
    fw_barrier_height
        Barrier height in forward direction in Hartree.
    bw_barrier_height
        Barrier height in backward direction in Hartree.
    temperature
        Temperature in Kelvin.
    imag_frequency
        Frequency in 1/s of the imaginary mode at the TS.

    Returns
    -------
    kappa
        Unitless tunneling correction according to Eckart.
    """
    kBT = KBAU * temperature
    hnu = PLANCKAU * imag_frequency * AU2SEC
    u = hnu / kBT

    def alpha(barr):
        return 2 * np.pi * barr / hnu

    alpha1 = alpha(fw_barrier_height)
    alpha2 = alpha(bw_barrier_height)
    kappa = tunl(alpha1, alpha2, u)
    return kappa


def get_rates(temperature, reactant_thermos, ts_thermo, product_thermos=None):
    G_TS = ts_thermo.G
    imag_wavenumber = ts_thermo.org_wavenumbers[0]
    assert imag_wavenumber < 0.0
    imag_frequency = (
        imag_wavenumber * C * 100
    )  # Convert from wavenumbers (1/cm) to (1/s)
    kappa_wigner = wigner_corr(temperature, imag_frequency)
    kappa_bell = bell_corr(temperature, imag_frequency)

    Gs_reactant = [thermo.G for thermo in reactant_thermos]
    G_reactant = sum(Gs_reactant)
    fw_barrier_height = G_TS - G_reactant
    fw_rate_eyring = eyring_rate(fw_barrier_height, temperature)

    if product_thermos:
        Gs_product = [thermo.G for thermo in product_thermos]
        G_product = sum(Gs_product)
        bw_barrier_height = G_TS - G_product
        bw_rate_eyring = eyring_rate(bw_barrier_height, temperature)
        kappa_eckart = eckart_corr(
            fw_barrier_height, bw_barrier_height, temperature, imag_frequency
        )
    else:
        kappa_eckart = None

    def make_rx_rates(from_, barrier, rate_eyring, kappa_eyring=1.0):
        kwargs = {}
        if kappa_wigner and not np.isnan(kappa_wigner):
            kwargs.update(
                {
                    "kappa_wigner": kappa_wigner,
                    "rate_wigner": kappa_wigner * rate_eyring,
                }
            )
        if kappa_bell and kappa_bell > 0.0:
            kwargs.update(
                {
                    "kappa_bell": kappa_bell,
                    "rate_bell": kappa_bell * rate_eyring,
                }
            )
        if kappa_eckart and not np.isnan(kappa_eckart):
            kwargs.update(
                {
                    "kappa_eckart": kappa_eckart,
                    "rate_eckart": kappa_eckart * rate_eyring,
                }
            )
        rx_rates = ReactionRates(
            from_=from_,
            barrier=barrier,
            barrier_si=barrier * AU2KJPERMOL,
            temperature=temperature,
            imag_wavenumber=imag_wavenumber,
            imag_frequency=imag_frequency,
            rate_eyring=rate_eyring,
            kappa_eyring=kappa_eyring,
            **kwargs,
        )
        return rx_rates

    reactant_rx_rates = make_rx_rates("Reactant(s)", fw_barrier_height, fw_rate_eyring)
    rx_rates = [
        reactant_rx_rates,
    ]
    if product_thermos:
        product_rx_rates = make_rx_rates(
            "Product(s)", bw_barrier_height, bw_rate_eyring
        )
        rx_rates.append(product_rx_rates)
    return rx_rates


def get_rates_for_geoms(temperature, reactant_geoms, ts_geom, product_geoms):
    def get_thermos(geoms):
        return [geom.get_thermoanalysis(T=temperature) for geom in geoms]

    reactant_thermos = get_thermos(reactant_geoms)
    ts_thermo = get_thermos((ts_geom,))[0]
    product_thermos = get_thermos(product_geoms)
    return get_rates(temperature, reactant_thermos, ts_thermo, product_thermos)


def get_rates_for_hessians(temperature, reactant_h5s, ts_h5, product_h5s):
    reactant_geoms = [geom_from_hessian(h5) for h5 in reactant_h5s]
    product_geoms = [geom_from_hessian(h5) for h5 in product_h5s]
    ts_geom = geom_from_hessian(ts_h5)
    rates = get_rates_for_geoms(temperature, reactant_geoms, ts_geom, product_geoms)
    return rates



from collections import OrderedDict
import re
import struct
import textwrap

from jinja2 import Template
import numpy as np

from pysisyphus.elem_data import KNOWN_ATOMS
from pysisyphus.constants import ANG2BOHR
from pysisyphus.Geometry import Geometry
from pysisyphus.helpers_pure import chunks, file_or_str
from pysisyphus.intcoords.setup_fast import find_bonds


AMINOS = (
    "ALA CYS ASP GLU PHE GLY HIS ILE LYS LEU"
    "MET ASN PRO GLN ARG SER THR VAL TRP TYR".split()
)


def get_parser(widths):
    fmt = " ".join("{}{}".format(width, "s") for width in widths)
    fieldstruct = struct.Struct(fmt)
    parse = lambda line: tuple(s.decode() for s in fieldstruct.unpack(line.encode()))
    return parse


STRIP_RE = re.compile(r"[\d\s]*")  # To remove numbers and whitespace
NAME_MAP = {
    "hb": "H",
    "he": "H",
    "hh": "H",
    "hd": "H",
    "hg": "H",
    "so": "Na",
}
FULL_NAME = {
    " sod",
    " cla",
    " cal",
}


def parse_atom_name(name):
    org_name = name
    assert len(name) == 4
    name = name.lower()
    # Cases like " SOD" require special handling. Sticking to the PDB specification (!)
    # and using only the first two characters would result in S (sulphur).
    use_full_name = name in FULL_NAME
    if not use_full_name:
        name = name[:2]
    stripped = STRIP_RE.sub("", name).lower()
    if use_full_name:
        stripped = stripped[:2]
    try:
        mapped = NAME_MAP[stripped]
    except KeyError:
        assert stripped in KNOWN_ATOMS, f"Could not parse atom name '{org_name}'"
        mapped = stripped
    return mapped.capitalize()


@file_or_str(".pdb")
def parse_pdb(text):
    atm_lines = list()
    for line in text.split("\n"):
        if line.startswith("HETATM") or line.startswith("ATOM"):
            atm_lines.append(line.strip())
            continue
        elif line.startswith("MASTER"):
            master_line = line.strip()

    """
        0  Record name
        1  serial
        2  name
        3  altLoc
        4  ResName
        5  chainID
        6  resSeq
        7  iCode
        8  x
        9  y
        10 z
        11 occupancy
        12 tempFactor
        13 element
        (14 charge), not parsed

        See https://stackoverflow.com/questions/4914008
    """
    atm_widths = (6, 6, 4, 1, 4, 1, 4, 1, 11, 8, 8, 6, 6, 12)
    atm_parse = get_parser(atm_widths)

    """
        0  Record name
        1  Num. of REMARK
        2  "0"
        3  Num. of HET
        4  Num. of HELIX
        5  Num. of SHEET
        6  deprecated
        7  Num. of SITE
        8  numXform
        9  Num. of atomic coordinates (HETATM + ATOMS)
        10 Num. of CONECT
        11 Num. of SEQRES
    """
    master_widths = (6, 9, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5)
    master_parse = get_parser(master_widths)
    try:
        master_fields = master_parse(master_line)
    except UnboundLocalError:
        master_fields = None

    atoms = list()
    coords = list()
    fragments = {}
    atom_map = dict()

    for i, line in enumerate(atm_lines):
        # Charge is not considered right now ...
        fields = atm_parse(f"{line[:78]: <78}")
        res_name = fields[4].strip()
        chain = fields[5].strip()
        res_seq = int(fields[6])
        frag = f"{chain}_{res_name}{res_seq}"

        xyz = np.array(fields[8:11], dtype=float)
        atom = fields[13].strip()
        if not atom.lower() in KNOWN_ATOMS:
            name = fields[2]
            atom = parse_atom_name(name)
        atoms.append(atom)
        coords.append(xyz)
        id_ = int(fields[1])
        atom_map[id_] = i

        frag = fragments.setdefault(frag, list())
        frag.append(i)

    # Verification using MASTER record, if present.
    if master_fields is not None:
        num_coord = int(master_fields[9])
        try:
            assert len(atoms) == num_coord
        except AssertionError as err:
            if len(atoms) < 99_999:  # See, e.g., 1HTQ
                raise err

    coords = np.array(coords, dtype=float) * ANG2BOHR
    return atoms, coords.flatten(), fragments, atom_map


def geom_from_pdb(fn, **kwargs):
    atoms, coords, fragments, _ = parse_pdb(fn)
    kwargs["fragments"] = fragments
    geom = Geometry(atoms, coords.flatten(), **kwargs)
    return geom


def get_conect_lines(atoms, coords):
    bonds = find_bonds(atoms, coords)
    # PDB indexing is 1-based
    bonds += 1
    # Bring in a suitable order for CONECT entries
    bonds.sort(axis=1)
    bonds = sorted(bonds, key=lambda row: row[0])
    conect = OrderedDict()
    for from_, to_ in bonds:
        conect.setdefault(from_, list()).append(to_)

    conect_lines = list()
    for from_, to_ in conect.items():
        to_iter = chunks(sorted(to_), 4)
        for to_chunk in to_iter:
            conect_lines.append([from_] + to_chunk)
    return conect_lines


def atoms_coords_to_pdb_str(atoms, coords, fragments=None, resname="", conect=True):
    coords3d = coords.reshape(-1, 3)
    coords3d_ang = coords3d / ANG2BOHR

    coord_fmt = "{: >8.3f}" * 3
    #            serial  name  altLoc      chainID     iCode
    #                                resName     resSeq
    hetatm_fmt = "HETATM{: >5d} {: >4}{: >1}{: >3}{: >1} {: >4d}{: >1}   "
    #      xyz
    hetatm_fmt += coord_fmt
    #      occupancy tempFactor            atom
    hetatm_fmt += "{: >6.2f}{: >6.2f}" + 10 * " " + "{: >2s}"
    ter_fmt = "TER   {: >5d}      {: >3s} {:1s}{: >4d}"

    if fragments is None:
        fragments = [
            range(len(atoms)),
        ]

    # Fixed for now
    altLoc = ""
    resName = resname
    chainID = ""
    iCode = ""
    occupancy = 1.0
    tempFactor = 0.0

    lines = list()
    serial = 1
    for resSeq, fragment in enumerate(fragments, 1):
        for id_ in fragment:
            name = atoms[id_]
            xyz = coords3d_ang[id_]
            line = hetatm_fmt.format(
                serial,
                name,
                altLoc,
                resName,
                chainID,
                resSeq,
                iCode,
                *xyz,
                occupancy,
                tempFactor,
                name,
            )
            lines.append(line)
            serial += 1
        lines.append(ter_fmt.format(serial, resName, chainID, resSeq))

    def fmt_conect_line(conect_line):
        return "CONECT" + ("{: >5d}" * len(conect_line)).format(*conect_line)

    if conect:
        conect_lines = get_conect_lines(atoms, coords)
    else:
        conect_lines = []
    conect_str = "\n".join([fmt_conect_line(cl) for cl in conect_lines])

    pdb_tpl = Template(
        textwrap.dedent(
            """\
    REMARK 1 Created by pysisyphus
    {%+ for line in lines -%}
    {{ line }}
    {%+ endfor -%}
    {{- conect_str }}
    END"""
        )
    )

    pdb_str = pdb_tpl.render(lines=lines, conect_str=conect_str)
    return pdb_str


def geom_to_pdb_str(geom, detect_fragments=False, **kwargs):
    fragments = None
    if detect_fragments:
        try:
            fragments = geom.internal.fragments
        except AttributeError:
            geom_ = geom.copy(coord_type="redund")
            fragments = geom_.internal.fragments
    pdb_str = atoms_coords_to_pdb_str(
        geom.atoms, geom.coords3d, fragments=fragments, **kwargs
    )
    return pdb_str



import numpy as np
from scipy.spatial.distance import pdist, squareform

from pysisyphus.calculators.Calculator import Calculator


class IDPPCalculator(Calculator):

    def __init__(self, target): 
        self.target = squareform(target)

        super().__init__(base_name="idpp")

    def get_forces(self, atoms, coords):
        coords_reshaped = coords.reshape((-1, 3))

        D = []
        for c in coords_reshaped:
            Di = coords_reshaped - c
            D.append(Di)
        D = np.array(D)

        curr_pdist = pdist(coords_reshaped)
        curr_square = squareform(curr_pdist)
        curr_diff = curr_square - self.target

        curr_square = curr_square + np.eye(curr_square.shape[0])

        # The bigger the differences 'curr_diff', the bigger the energy.
        # The smaller the current distances 'current_pdist', the bigger
        # the energy.
        energy = 0.5 * (curr_diff**2 / curr_square**4).sum()

        # Adapted from ASE IDPP calculator
        # https://gitlab.com/ase/ase/blob/master/ase/neb.py, GPL2
        forces = -2 * ((curr_diff *
                       (1 - 2 * curr_diff / curr_square) /
                        curr_square**5)[...,np.newaxis] * D).sum(0)

        results = {
            "energy" : energy,
            "forces": forces.flatten()
        }
        return results

    def __str__(self):
        return "IDPP calculator"





import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.derivatives import dq_oop, d2q_oop
from pysisyphus.linalg import cross3, norm3


class OutOfPlane(Primitive):
    """
    [1] https://doi.org/10.1002/(SICI)1096-987X(19990730)20:10<1067::AID-JCC9>3.0.CO;2-V
        Lee, 1999
    """

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        raise Exception("Not yet implemented!")

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        """
              P
            / | \
           /  |  \
          u'  v'  w'
         /    |    \
        m     n     o
        """
        # p is apex
        m, n, o, p = indices

        u_dash = coords3d[m] - coords3d[p]
        v_dash = coords3d[n] - coords3d[p]
        w_dash = coords3d[o] - coords3d[p]

        u = u_dash / norm3(u_dash)
        v = v_dash / norm3(v_dash)
        w = w_dash / norm3(w_dash)

        z_dash = cross3(u, v) + cross3(v, w) + cross3(w, u)
        z = z_dash / norm3(z_dash)

        oop_coord = z.dot(u)

        if gradient:
            grad = dq_oop(*coords3d[m], *coords3d[n], *coords3d[o], *coords3d[p])
            grad = grad.reshape(4, 3)
            row = np.zeros_like(coords3d)
            row[m, :] = grad[0]
            row[n, :] = grad[1]
            row[o, :] = grad[2]
            row[p, :] = grad[3]
            row = row.flatten()
            return oop_coord, row

        return oop_coord

    @staticmethod
    def _jacobian(coords3d, indices):
        return d2q_oop(*coords3d[indices].flatten())



"""
Molecular integrals over Gaussian basis functions generated by sympleints.
See https://github.com/eljost/sympleints for more information.

sympleints version: 0.1.dev79+g63f1ef8.d20230515
symppy version: 1.10.1

sympleints was executed with the following arguments:
	lmax = 4
	lauxmax = 6
	write = False
	out_dir = devel_ints
	keys = ['~2c2e', '~3c2e_sph']
	sph = False
	opt_basic = True
	normalize = cgto
"""

import numpy


def self_ovlp3d_00(ax, da, A, bx, db, B):
    """Cartesian 3D (ss) self overlap.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((1, 1), dtype=float)

    x0 = (ax + bx) ** (-1.0)

    # 1 item(s)
    result[0, 0] = numpy.sum(5.568327996831708 * da * db * x0**1.5)
    return result


def self_ovlp3d_11(ax, da, A, bx, db, B):
    """Cartesian 3D (pp) self overlap.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((3, 3), dtype=float)

    x0 = ax + bx
    x1 = 2.784163998415854 * da * db * numpy.sqrt(x0 ** (-1.0)) / x0**2

    # 3 item(s)
    result[0, 0] = numpy.sum(x1)
    result[0, 1] = numpy.sum(x1)
    result[0, 2] = numpy.sum(x1)
    return result


def self_ovlp3d_22(ax, da, A, bx, db, B):
    """Cartesian 3D (dd) self overlap.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((6, 6), dtype=float)

    x0 = ax + bx
    x1 = x0 ** (-1.0)
    x2 = 1.392081999207927 * da * db
    x3 = numpy.sqrt(x1) * x2 / x0**3
    x4 = x1**1.5 * x2 / x0**2

    # 6 item(s)
    result[0, 0] = numpy.sum(x3)
    result[0, 1] = numpy.sum(x4)
    result[0, 2] = numpy.sum(x4)
    result[0, 3] = numpy.sum(x3)
    result[0, 4] = numpy.sum(x4)
    result[0, 5] = numpy.sum(x3)
    return result


def self_ovlp3d_33(ax, da, A, bx, db, B):
    """Cartesian 3D (ff) self overlap.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((10, 10), dtype=float)

    x0 = ax + bx
    x1 = x0 ** (-1.0)
    x2 = 0.6960409996039635 * da * db
    x3 = numpy.sqrt(x1) * x2 / x0**4
    x4 = x1**1.5 * x2 / x0**3

    # 10 item(s)
    result[0, 0] = numpy.sum(x3)
    result[0, 1] = numpy.sum(x4)
    result[0, 2] = numpy.sum(x4)
    result[0, 3] = numpy.sum(x4)
    result[0, 4] = numpy.sum(x4)
    result[0, 5] = numpy.sum(x4)
    result[0, 6] = numpy.sum(x3)
    result[0, 7] = numpy.sum(x4)
    result[0, 8] = numpy.sum(x4)
    result[0, 9] = numpy.sum(x3)
    return result


def self_ovlp3d_44(ax, da, A, bx, db, B):
    """Cartesian 3D (gg) self overlap.

    Generated code; DO NOT modify by hand!"""

    result = numpy.zeros((15, 15), dtype=float)

    x0 = 0.5 / (ax + bx)
    x1 = (ax + bx) ** (-1.0)
    x2 = (x1 * (ax + bx) - 1.0) * A[0]
    x3 = 1.772453850905516 * numpy.sqrt(x1)
    x4 = x2**2 * x3
    x5 = x0 * x3
    x6 = 3.0 * x5
    x7 = x0 * (3.0 * x4 + x6)
    x8 = x4 + x5
    x9 = x2 * x8
    x10 = x2 * x5
    x11 = x2 * (2.0 * x10 + x9)
    x12 = 5.0 * x0 * (x11 + x7)
    x13 = 4.0 * x0 * (2.0 * x10 + x9)
    x14 = x11 + x7
    x15 = x14 * x2
    x16 = x2 * (x13 + x15)
    x17 = x12 + x16
    x18 = da * db
    x19 = 0.02991993003418851 * x1 * x18
    x20 = (x1 * (ax + bx) - 1.0) * A[1]
    x21 = x20**2 * x3
    x22 = x21 + x5
    x23 = x18 * x22
    x24 = 0.06666666666666667 * x3
    x25 = x17 * x24
    x26 = (x1 * (ax + bx) - 1.0) * A[2]
    x27 = x26**2 * x3
    x28 = x27 + x5
    x29 = x18 * x28
    x30 = x0 * (3.0 * x21 + x6)
    x31 = x20 * x22
    x32 = 2.0 * x5
    x33 = x20 * (x20 * x32 + x31)
    x34 = x30 + x33
    x35 = x18 * x34
    x36 = 0.1111111111111111 * x14 * x3
    x37 = 0.3333333333333333 * x28
    x38 = x0 * (3.0 * x27 + x6)
    x39 = x26 * x28
    x40 = x26 * (x26 * x32 + x39)
    x41 = x38 + x40
    x42 = 5.0 * x0 * (x30 + x33)
    x43 = 8.0 * x5
    x44 = x0 * (x20 * x43 + 4.0 * x31)
    x45 = x20 * x34
    x46 = x20 * (x44 + x45)
    x47 = x42 + x46
    x48 = x24 * x47
    x49 = x18 * x8
    x50 = 5.0 * x0 * (x38 + x40)
    x51 = x0 * (x26 * x43 + 4.0 * x39)
    x52 = x26 * x41
    x53 = x26 * (x51 + x52)
    x54 = x50 + x53
    x55 = x24 * x54

    # 15 item(s)
    result[0, 0] = numpy.sum(
        x19 * (7.0 * x0 * (x12 + x16) + x2 * (6.0 * x0 * (x13 + x15) + x17 * x2))
    )
    result[0, 1] = numpy.sum(x23 * x25)
    result[0, 2] = numpy.sum(x25 * x29)
    result[0, 3] = numpy.sum(x35 * x36)
    result[0, 4] = numpy.sum(x14 * x23 * x37)
    result[0, 5] = numpy.sum(x18 * x36 * x41)
    result[0, 6] = numpy.sum(x48 * x49)
    result[0, 7] = numpy.sum(x35 * x37 * x8)
    result[0, 8] = numpy.sum(0.3333333333333333 * x23 * x41 * x8)
    result[0, 9] = numpy.sum(x49 * x55)
    result[0, 10] = numpy.sum(
        x19 * (7.0 * x0 * (x42 + x46) + x20 * (6.0 * x0 * (x44 + x45) + x20 * x47))
    )
    result[0, 11] = numpy.sum(x29 * x48)
    result[0, 12] = numpy.sum(0.1111111111111111 * x3 * x35 * x41)
    result[0, 13] = numpy.sum(x23 * x55)
    result[0, 14] = numpy.sum(
        x19 * (7.0 * x0 * (x50 + x53) + x26 * (6.0 * x0 * (x51 + x52) + x26 * x54))
    )
    return result


self_ovlp3d = {
    (0, 0): self_ovlp3d_00,
    (1, 1): self_ovlp3d_11,
    (2, 2): self_ovlp3d_22,
    (3, 3): self_ovlp3d_33,
    (4, 4): self_ovlp3d_44,
}



from math import sqrt

import numpy as np


class NormalMode:
    """See http://gaussian.com/vib/"""

    def __init__(self, l, masses):
        """NormalMode class.

        Cartesian displacements are normalized to 1.

        Parameters
        ----------
        l : np.array
            Cartesian, non-mass-weighted displacements.
        masses : np.array
            Atomic masses.
        """

        self.l = np.array(l.flatten())
        self.l /= np.linalg.norm(l)
        self.masses = masses
        assert self.l.shape == self.masses.shape

    def __len__(self):
        return self.l.size

    @property
    def red_mass(self):
        return 1 / np.sum(np.square(self.l_mw) / self.masses)

    def mw_norm_for_norm(self, norm=0.01):
        return norm * sqrt(self.red_mass)

    @property
    def l_mw(self):
        l_mw = self.l * np.sqrt(self.masses)
        return l_mw / np.linalg.norm(l_mw)



import numpy as np

from pysisyphus.intcoords.Torsion import Torsion


class DummyImproper(Torsion):
    def __init__(self, indices, *args, fix_inner=True, **kwargs):
        self.fix_inner = fix_inner
        kwargs["calc_kwargs"] = ("fix_inner",)
        super().__init__(indices, *args, **kwargs)
        self.log("DummyImproper is never checked for collinear atoms!")

    @staticmethod
    def _get_dummy_coords(coords3d, indices, r=1.889):
        r"""
    dummy -> D\
    atomy    | \
             |  \
         A---B---C
        """
        a, b, c = indices
        # Center
        B = coords3d[b]
        # Bond, pointing away from a to c
        AC_ = coords3d[c] - coords3d[a]
        AC = AC_ / np.linalg.norm(AC_)

        w = DummyImproper._get_cross_vec(coords3d, indices).flatten()
        dummy_vec = (np.eye(3) - np.outer(AC, AC)) @ w
        dummy_coords = B + r * dummy_vec
        return dummy_coords

    @staticmethod
    def get_coords3d_and_indices_ext(coords3d, indices):
        dummy_coords = DummyImproper._get_dummy_coords(coords3d, indices)
        dummy_ind = len(coords3d)
        coords3d_ext = np.zeros((len(coords3d) + 1, 3))
        coords3d_ext[:dummy_ind] = coords3d
        coords3d_ext[dummy_ind] = dummy_coords
        a, b, c = indices
        indices_ext = a, b, dummy_ind, c
        return coords3d_ext, indices_ext

    @staticmethod
    def _weight(*args, **kwargs):
        return 1

    @staticmethod
    def _calculate(coords3d, indices, gradient=False, fix_inner=False):
        coords3d_ext, indices_ext = DummyImproper.get_coords3d_and_indices_ext(
            coords3d, indices
        )

        results = Torsion._calculate(coords3d_ext, indices_ext, gradient=gradient)
        if gradient:
            val, grad = results

            # import pdb; pdb.set_trace()  # fmt: skip
            # Remove entries that belong to the dummy atom.
            grad = grad[:-3]  # .reshape[](-1, 3)[[0, 1, 3]]

            # Zero out contributions of the inner two atoms in the torsion.
            # So basically only the atom at the first index moves.
            #
            # This usually degrades optimization convergence.
            # if fix_inner:
            # grad.reshape(-1, 3)[indices_ext[1:3]] = 0.0
            return val, grad.flatten()
        else:
            return results



from math import sin

import numpy as np

from pysisyphus.intcoords.Primitive import Primitive
from pysisyphus.intcoords.derivatives import d2q_a
from pysisyphus.linalg import cross3, norm3


class Bend(Primitive):

    @staticmethod
    def _weight(atoms, coords3d, indices, f_damping):
        m, o, n = indices
        rho_mo = Bend.rho(atoms, coords3d, (m, o))
        rho_on = Bend.rho(atoms, coords3d, (o, n))
        rad = Bend._calculate(coords3d, indices)
        return (rho_mo * rho_on)**0.5 * (f_damping + (1-f_damping)*sin(rad))

    @staticmethod
    def _calculate(coords3d, indices, gradient=False):
        m, o, n = indices
        u_dash = coords3d[m] - coords3d[o]
        v_dash = coords3d[n] - coords3d[o]
        u_norm = norm3(u_dash)
        v_norm = norm3(v_dash)
        u = u_dash / u_norm
        v = v_dash / v_norm

        udv = max(-1.0, min(1.0, u.dot(v)))
        angle_rad = np.arccos(udv)

        if gradient:
            cross_vec1 = ( 1, -1, 1)
            cross_vec2 = (-1,  1, 1)

            # Determine second vector for the cross product, to get an
            # orthogonal direction. Eq. (24) in [1]
            uv_parallel = Bend.parallel(u, v)
            if not uv_parallel:
                cross_vec = v
            elif not Bend.parallel(u, cross_vec1):
                cross_vec = cross_vec1
            else:
                cross_vec = cross_vec2

            w_dash = cross3(u, cross_vec)
            w = w_dash / norm3(w_dash)

            uxw = cross3(u, w)
            wxv = cross3(w, v)

            row = np.zeros_like(coords3d)
            #                  |  m  |  n  |  o  |
            # -----------------------------------
            # sign_factor(amo) |  1  |  0  | -1  | first_term
            # sign_factor(ano) |  0  |  1  | -1  | second_term
            first_term = uxw / u_norm
            second_term = wxv / v_norm
            row[m,:] = first_term
            row[o,:] = -first_term - second_term
            row[n,:] = second_term
            row = row.flatten()
            return angle_rad, row
        return angle_rad

    @staticmethod
    def _jacobian(coords3d, indices):
        return d2q_a(*coords3d[indices].flatten())



"""CLI bool-argument normalization helpers.

This module keeps backward compatibility for legacy value-style boolean flags:
`--flag True/False`.

Also provides option-name case normalization: ``normalize_argv_option_names``
lowercases long option names (``--Add-LinkH`` → ``--add-linkh``) so that flag
input is case-insensitive while the canonical form stays lowercase.
"""

from __future__ import annotations


_BOOL_TRUE_LITERALS = {"1", "true", "t", "yes", "y", "on"}
_BOOL_FALSE_LITERALS = {"0", "false", "f", "no", "n", "off"}


def normalize_argv_option_names(args: list[str]) -> list[str]:
    """Lowercase long option names, leaving values unchanged.

    ``--Add-LinkH`` → ``--add-linkh``
    ``--include-H2O=True`` → ``--include-h2o=True``
    Short flags (``-v``) and positional arguments are passed through unchanged.
    """
    result: list[str] = []
    for token in args:
        if token.startswith("--"):
            name, sep, value = token.partition("=")
            result.append(name.lower() + sep + value)
        else:
            result.append(token)
    return result


def _parse_bool_literal(raw: str) -> bool | None:
    token = raw.strip().lower()
    if token in _BOOL_TRUE_LITERALS:
        return True
    if token in _BOOL_FALSE_LITERALS:
        return False
    return None


def _toggle_negative_name(
    command: str,
    positive_name: str,
    toggle_negative_aliases: dict[str, dict[str, str]],
) -> str:
    aliases = toggle_negative_aliases.get(command)
    if aliases and positive_name in aliases:
        return aliases[positive_name]
    return f"--no-{positive_name[2:]}"


def _read_bool_literal(
    args: list[str], index: int, sep: str, inline_value: str
) -> tuple[bool | None, int, bool]:
    """Return (parsed_bool, next_index, explicit_literal_used)."""
    if sep:
        return _parse_bool_literal(inline_value), index + 1, True
    if index + 1 < len(args):
        parsed_next = _parse_bool_literal(args[index + 1])
        if parsed_next is not None:
            return parsed_next, index + 2, True
    return None, index + 1, False


def _build_negative_to_positive_toggle_map(
    toggle_options: frozenset[str],
    toggle_negative_aliases: dict[str, dict[str, str]],
    command: str,
) -> dict[str, str]:
    negative_to_positive: dict[str, str] = {}
    for positive_name in toggle_options:
        if not positive_name.startswith("--"):
            continue
        canonical_negative = _toggle_negative_name(
            command, positive_name, toggle_negative_aliases
        )
        negative_to_positive[canonical_negative] = positive_name
        synthetic_no_name = f"--no-{positive_name[2:]}"
        negative_to_positive.setdefault(synthetic_no_name, positive_name)
    return negative_to_positive


def _build_negative_to_positive_single_flag_map(
    single_flag_options: frozenset[str],
) -> dict[str, str]:
    negative_to_positive: dict[str, str] = {}
    for positive_name in single_flag_options:
        if not positive_name.startswith("--"):
            continue
        negative_to_positive[f"--no-{positive_name[2:]}"] = positive_name
    return negative_to_positive


def normalize_bool_argv(
    args: list[str],
    bool_value_options_by_command: dict[str, frozenset[str]],
    bool_toggle_options_by_command: dict[str, frozenset[str]],
    toggle_negative_aliases: dict[str, dict[str, str]],
    bool_single_flag_options_by_command: dict[str, frozenset[str]] | None = None,
) -> tuple[list[str], bool]:
    """Normalize CLI argv booleans and return (normalized_args, legacy_syntax_used)."""
    if not args:
        return args, False

    command = args[0]
    bool_value_options = bool_value_options_by_command.get(command, frozenset())
    bool_toggle_options = bool_toggle_options_by_command.get(command, frozenset())
    bool_single_flag_options = (
        bool_single_flag_options_by_command or {}
    ).get(command, frozenset())

    if not bool_value_options and not bool_toggle_options and not bool_single_flag_options:
        return args, False

    toggle_negative_to_positive = _build_negative_to_positive_toggle_map(
        bool_toggle_options, toggle_negative_aliases, command
    )
    single_negative_to_positive = _build_negative_to_positive_single_flag_map(
        bool_single_flag_options
    )

    normalized: list[str] = [command]
    legacy_used = False
    i = 1
    while i < len(args):
        token = args[i]

        if token == "--":
            normalized.extend(args[i:])
            break

        if not token.startswith("--"):
            normalized.append(token)
            i += 1
            continue

        name, sep, inline_value = token.partition("=")

        if name in toggle_negative_to_positive:
            positive_name = toggle_negative_to_positive[name]
            canonical_negative = _toggle_negative_name(
                command, positive_name, toggle_negative_aliases
            )
            parsed, next_i, explicit_literal = _read_bool_literal(
                args, i, sep, inline_value
            )
            if explicit_literal:
                if parsed is None:
                    normalized.append(token)
                else:
                    legacy_used = True
                    normalized.append(canonical_negative if parsed else positive_name)
                i = next_i
                continue

            normalized.append(canonical_negative)
            i += 1
            continue

        if name in single_negative_to_positive:
            positive_name = single_negative_to_positive[name]
            parsed, next_i, explicit_literal = _read_bool_literal(
                args, i, sep, inline_value
            )
            if explicit_literal:
                if parsed is None:
                    normalized.append(token)
                else:
                    legacy_used = True
                    if not parsed:
                        normalized.append(positive_name)
                    else:
                        normalized.append(name)
                i = next_i
                continue

            i += 1
            continue

        if name.startswith("--no-"):
            positive_name = "--" + name[5:]
            if sep == "" and positive_name in bool_value_options:
                normalized.extend([positive_name, "False"])
                i += 1
                continue
            normalized.append(token)
            i += 1
            continue

        if name in bool_value_options:
            parsed, next_i, explicit_literal = _read_bool_literal(args, i, sep, inline_value)
            if explicit_literal:
                if parsed is not None:
                    legacy_used = True
                    normalized.extend([name, "True" if parsed else "False"])
                else:
                    normalized.append(token)
                i = next_i
                continue

            normalized.extend([name, "True"])
            i += 1
            continue

        if name in bool_toggle_options:
            parsed, next_i, explicit_literal = _read_bool_literal(args, i, sep, inline_value)
            if explicit_literal:
                if parsed is not None:
                    legacy_used = True
                    normalized.append(
                        name if parsed else _toggle_negative_name(
                            command, name, toggle_negative_aliases
                        )
                    )
                else:
                    normalized.append(token)
                i = next_i
                continue

            normalized.append(name)
            i += 1
            continue

        if name in bool_single_flag_options:
            parsed, next_i, explicit_literal = _read_bool_literal(args, i, sep, inline_value)
            if explicit_literal:
                if parsed is not None:
                    legacy_used = True
                    if parsed:
                        normalized.append(name)
                else:
                    normalized.append(token)
                i = next_i
                continue

            normalized.append(name)
            i += 1
            continue

        normalized.append(token)
        i += 1

    return normalized, legacy_used



from pysisyphus.calculators.AnaPotBase import AnaPotBase
from pysisyphus.calculators.LEPSExpr import LEPSExpr


class LEPSBase(AnaPotBase):

    def __init__(self, pot_type="leps"):
        leps_expr = LEPSExpr()
        V_expr, xlim, ylim, levels = leps_expr.get_expr(pot_type)
        V_str = str(V_expr)

        super().__init__(V_str=V_str, xlim=xlim, ylim=ylim, levels=levels)

    def __str__(self):
        return "LEPSBase calculator"


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    choices = "leps harmonic tot dimer".split()
    for c in choices:
        lp = LEPSBase(pot_type=c)
        lp.plot()
        plt.show()



#!/usr/bin/env python3
"""Lightweight smoke tests for commands embedded in docs markdown files."""

from __future__ import annotations

import argparse
import os
import shlex
import subprocess
import sys
import tempfile
from pathlib import Path

from click.testing import CliRunner

REPO_ROOT = Path(__file__).resolve().parents[1]
DOCS_ROOT = REPO_ROOT / "docs"
TOOL_NAME = "pdb2reaction"
CLI_MODULE = "pdb2reaction"
DOCS_SMOKE_COMMAND_TIMEOUT_SEC = float(os.environ.get("DOCS_SMOKE_COMMAND_TIMEOUT_SEC", "120"))

sys.path.insert(0, str(REPO_ROOT))

from pdb2reaction.cli import cli as root_cli  # noqa: E402


_CODE_LANGS = {"", "bash", "sh", "shell", "console"}
_ALL_ONLY_PATH_EXTS = {".pdb", ".xyz", ".gjf", ".yaml", ".yml", ".json"}


def _extract_commands_from_block(lines: list[str]) -> list[str]:
    commands: list[str] = []
    current = ""
    for raw in lines:
        line = raw.strip()
        if not line or line.startswith("#"):
            continue
        if " #" in line:
            line = line.split(" #", 1)[0].rstrip()
            if not line:
                continue
        if line.startswith("$"):
            line = line[1:].strip()
        current = f"{current} {line}".strip() if current else line
        if current.endswith("\\"):
            current = current[:-1].rstrip()
            continue
        commands.append(current)
        current = ""
    if current:
        commands.append(current)
    filtered: list[str] = []
    for cmd in commands:
        if not cmd.startswith(TOOL_NAME):
            continue
        if any(mark in cmd for mark in ("<", ">", "[", "]")):
            continue
        filtered.append(cmd)
    return filtered


def _extract_docs_commands() -> list[str]:
    commands: list[str] = []
    for path in sorted(DOCS_ROOT.rglob("*.md")):
        lines = path.read_text(encoding="utf-8").splitlines()
        in_fence = False
        fence_lang = ""
        block: list[str] = []
        for line in lines:
            stripped = line.strip()
            if stripped.startswith("```"):
                marker = stripped[3:].strip().lower()
                if not in_fence:
                    in_fence = True
                    fence_lang = marker
                    block = []
                else:
                    if fence_lang in _CODE_LANGS:
                        commands.extend(_extract_commands_from_block(block))
                    in_fence = False
                    fence_lang = ""
                    block = []
                continue
            if in_fence:
                block.append(line)
    return commands


def _subcommand_from_tokens(tokens: list[str]) -> str:
    if len(tokens) < 2 or tokens[1].startswith("-"):
        return "all"
    return tokens[1]


def _prepare_fixture_files(tmp: Path) -> dict[str, Path]:
    pdb_text = (
        "ATOM      1  C   LIG A   1       0.000   0.000   0.000  1.00  0.00           C\n"
        "END\n"
    )
    r_pdb = tmp / "R.pdb"
    p_pdb = tmp / "P.pdb"
    xyz = tmp / "input.xyz"
    gjf = tmp / "input.gjf"
    cfg = tmp / "config.yaml"
    out_dir = tmp / "result_all"

    r_pdb.write_text(pdb_text, encoding="utf-8")
    p_pdb.write_text(pdb_text, encoding="utf-8")
    xyz.write_text("1\n\nC 0.0 0.0 0.0\n", encoding="utf-8")
    gjf.write_text("%chk=test\n#p hf/3-21g\n\nTitle\n\n0 1\nC 0.0 0.0 0.0\n\n", encoding="utf-8")
    cfg.write_text("extract:\n  radius: 2.6\n", encoding="utf-8")
    out_dir.mkdir(parents=True, exist_ok=True)

    return {
        "r_pdb": r_pdb,
        "p_pdb": p_pdb,
        "xyz": xyz,
        "gjf": gjf,
        "config": cfg,
        "out_dir": out_dir,
    }


def _sanitize_all_args(args: list[str], fixture: dict[str, Path]) -> list[str]:
    out: list[str] = []
    saw_input = False
    saw_dry_run = False
    saw_center = False
    i = 0
    while i < len(args):
        tok = args[i]
        if tok in {"--version", "-h", "--help"}:
            i += 1
            continue
        if tok in {"-c", "--center"}:
            saw_center = True
            out.append(tok)
            if i + 1 < len(args):
                out.append(args[i + 1])
            i += 2
            continue
        if tok in {"-i", "--input"}:
            saw_input = True
            out.extend([tok, str(fixture["r_pdb"]), str(fixture["p_pdb"])])
            i += 1
            while i < len(args) and not args[i].startswith("-"):
                i += 1
            continue
        if tok == "--config":
            out.extend([tok, str(fixture["config"])])
            i += 2
            continue
        if tok == "--out-dir":
            out.extend([tok, str(fixture["out_dir"])])
            i += 2
            continue
        if tok == "--dry-run":
            saw_dry_run = True
            out.append(tok)
            i += 1
            continue
        if tok == "--no-dry-run":
            saw_dry_run = True
            out.append("--dry-run")
            i += 1
            continue
        if (not tok.startswith("-")) and Path(tok).suffix.lower() in _ALL_ONLY_PATH_EXTS:
            ext = Path(tok).suffix.lower()
            if ext == ".pdb":
                out.append(str(fixture["r_pdb"]))
            elif ext == ".xyz":
                out.append(str(fixture["xyz"]))
            elif ext == ".gjf":
                out.append(str(fixture["gjf"]))
            else:
                out.append(str(fixture["config"]))
            i += 1
            continue

        out.append(tok)
        i += 1

    if not saw_input:
        out.extend(["-i", str(fixture["r_pdb"]), str(fixture["p_pdb"])])
    if not saw_center:
        out.extend(["-c", "LIG"])
    if not saw_dry_run:
        out.append("--dry-run")
    if "--out-dir" not in out:
        out.extend(["--out-dir", str(fixture["out_dir"])])
    return out


def _run_help_smoke(commands: list[str]) -> None:
    runner = CliRunner()
    subcommands = sorted({_subcommand_from_tokens(shlex.split(cmd)) for cmd in commands})
    for subcmd in subcommands:
        result = runner.invoke(root_cli, [subcmd, "--help"], catch_exceptions=False)
        if result.exit_code != 0:
            raise RuntimeError(
                f"[help-smoke] failed for '{TOOL_NAME} {subcmd} --help':\n{result.output}"
            )
    print(f"[help-smoke] validated {len(subcommands)} subcommands from docs.")


def _run_all_dry_run_smoke(commands: list[str]) -> None:
    try:
        probe = subprocess.run(
            [sys.executable, "-m", CLI_MODULE, "all", "--help"],
            cwd=REPO_ROOT,
            text=True,
            capture_output=True,
            timeout=DOCS_SMOKE_COMMAND_TIMEOUT_SEC,
        )
    except subprocess.TimeoutExpired as exc:
        raise RuntimeError(
            "[dry-run-smoke] timeout while probing availability for "
            f"'{TOOL_NAME} all --help' ({DOCS_SMOKE_COMMAND_TIMEOUT_SEC:g}s)."
        ) from exc
    probe_output = f"{probe.stdout}\n{probe.stderr}"
    if "Command 'all' is unavailable" in probe_output or "Missing dependency:" in probe_output:
        print("[dry-run-smoke] skipped: 'all' command is unavailable in this environment.")
        return

    all_cmds: set[str] = set()
    for cmd in commands:
        tokens = shlex.split(cmd)
        if not tokens or tokens[0] != TOOL_NAME:
            continue
        if len(tokens) >= 2 and tokens[1] == "all":
            all_cmds.add(cmd)
            continue
        if len(tokens) >= 2 and tokens[1].startswith("-"):
            if any(tok in {"-i", "--input"} for tok in tokens[1:]):
                all_cmds.add(cmd)
    all_cmds = sorted(all_cmds)
    if not all_cmds:
        raise RuntimeError("No 'all' command examples found in docs.")

    with tempfile.TemporaryDirectory(prefix=f"{TOOL_NAME}_docs_smoke_") as tmpdir:
        fixture = _prepare_fixture_files(Path(tmpdir))
        for raw in all_cmds:
            tokens = shlex.split(raw)
            if not tokens or tokens[0] != TOOL_NAME:
                continue
            args = tokens[1:]
            if not args or args[0].startswith("-"):
                args = ["all", *args]
            if args[0] != "all":
                continue
            dry_args = _sanitize_all_args(args, fixture)
            try:
                completed = subprocess.run(
                    [sys.executable, "-m", CLI_MODULE, *dry_args],
                    cwd=REPO_ROOT,
                    text=True,
                    capture_output=True,
                    timeout=DOCS_SMOKE_COMMAND_TIMEOUT_SEC,
                )
            except subprocess.TimeoutExpired as exc:
                raise RuntimeError(
                    f"[dry-run-smoke] timeout for docs command ({DOCS_SMOKE_COMMAND_TIMEOUT_SEC:g}s):\n"
                    f"  {raw}\n"
                    f"sanitized args:\n"
                    f"  {TOOL_NAME} {' '.join(dry_args)}"
                ) from exc
            if completed.returncode != 0:
                raise RuntimeError(
                    f"[dry-run-smoke] failed for docs command:\n  {raw}\n"
                    f"sanitized args:\n  {TOOL_NAME} {' '.join(dry_args)}\n\n"
                    f"stdout:\n{completed.stdout}\n\nstderr:\n{completed.stderr}"
                )
    print(
        f"[dry-run-smoke] validated {len(all_cmds)} docs examples "
        f"(timeout={DOCS_SMOKE_COMMAND_TIMEOUT_SEC:g}s)."
    )


def main() -> int:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.parse_args()

    commands = _extract_docs_commands()
    if not commands:
        raise RuntimeError("No commands were extracted from docs markdown code fences.")

    _run_help_smoke(commands)
    _run_all_dry_run_smoke(commands)
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from math import cos, sin, sqrt
from typing import Callable, Literal, Optional, Tuple

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation
from scipy.linalg.lapack import dpstrf


def gram_schmidt(vecs, thresh=1e-8):
    def proj(v1, v2):
        return v1.dot(v2) / v1.dot(v1)

    ortho = [
        vecs[0],
    ]
    for v1 in vecs[1:]:
        tmp = v1.copy()
        for v2 in ortho:
            tmp -= proj(v2, v1) * v2
        norm = np.linalg.norm(tmp)
        # Don't append linear dependent vectors, as their norm will be
        # near zero. Renormalizing them to unity would lead to numerical
        # garbage and to erronous results later on, when we orthgonalize
        # against this 'arbitrary' vector.
        if norm <= thresh:
            continue
        ortho.append(tmp / norm)
    return np.array(ortho)


def orthogonalize_against(mat, vecs, max_cycles=5, thresh=1e-10):
    """Orthogonalize rows of 'mat' against rows in 'vecs'

    Returns a (modified) copy of mat.
    """

    omat = mat.copy()
    for _ in range(max_cycles):
        max_overlap = 0.0
        for row in omat:
            for ovec in vecs:
                row -= ovec.dot(row) * ovec

            # Remaining overlap
            overlap = np.sum(np.dot(vecs, row))
            max_overlap = max(overlap, max_overlap)

        if max_overlap < thresh:
            break
    else:
        raise Exception(f"Orthogonalization did not succeed in {max_cycles} cycles!")

    return omat


def perp_comp(vec, along):
    """Return the perpendicular component of vec along along."""
    return vec - vec.dot(along) * along


def make_unit_vec(vec1, vec2):
    """Return unit vector pointing from vec2 to vec1."""
    diff = vec1 - vec2
    return diff / np.linalg.norm(diff)


def svd_inv(array, thresh, hermitian=False):
    U, S, Vt = np.linalg.svd(array, hermitian=hermitian)
    keep = S > thresh
    S_inv = np.zeros_like(S)
    S_inv[keep] = 1 / S[keep]
    return Vt.T.dot(np.diag(S_inv)).dot(U.T)


def get_rot_mat(abc=None):
    # Euler angles
    if abc is None:
        abc = np.random.rand(3) * np.pi * 2
    a, b, c = abc
    R = np.array(
        (
            (
                cos(a) * cos(b) * cos(c) - sin(a) * sin(c),
                -cos(a) * cos(b) * sin(c) - sin(a) * cos(c),
                cos(a) * sin(b),
            ),
            (
                sin(a) * cos(b) * cos(c) + cos(a) * sin(c),
                -sin(a) * cos(b) * sin(c) + cos(a) * cos(c),
                sin(a) * sin(b),
            ),
            (-sin(b) * cos(c), sin(b) * sin(c), cos(b)),
        )
    )
    return R


def get_rot_mat_for_coords(coords3d_1, coords3d_2):
    coords3d_1 = coords3d_1.copy()
    coords3d_2 = coords3d_2.copy()
    centroid_1 = coords3d_1.mean(axis=0)
    centroid_2 = coords3d_2.mean(axis=0)
    coords3d_1 -= centroid_1[None, :]
    coords3d_2 -= centroid_2[None, :]

    tmp = coords3d_2.T.dot(coords3d_1)
    U, W, Vt = np.linalg.svd(tmp)
    rot_mat = U.dot(Vt)
    if np.linalg.det(rot_mat) < 0:
        U[:, -1] *= -1
        rot_mat = U.dot(Vt)
    return coords3d_1, coords3d_2, rot_mat


def eigvec_grad(w, v, ind, mat_grad):
    """Gradient of 'ind'-th eigenvector.

    dv_i / dx_i = (w_i*I - mat)⁻¹ dmat/dx_i v_i
    """
    eigval = w[ind]
    eigvec = v[:, ind]

    w_diff = eigval - w
    w_inv = np.divide(
        1.0, w_diff, out=np.zeros_like(w_diff).astype(float), where=w_diff != 0.0
    )
    assert np.isfinite(w_inv).all()
    pinv = v.dot(np.diag(w_inv)).dot(v.T)
    pinv.dot(mat_grad)

    wh = pinv.dot(mat_grad).dot(eigvec)
    return wh


def cross3(a, b):
    """10x as fast as np.cross for two 1d arrays of size 3."""
    return np.array(
        (
            a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0],
        )
    )


def norm3(a):
    """5x as fas as np.linalg.norm for a 1d array of size 3."""
    return sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2])


def finite_difference_hessian(
    coords: NDArray[float],
    grad_func: Callable[[NDArray[float]], NDArray[float]],
    step_size: float = 1e-2,
    acc: Literal[2, 4] = 2,
    callback: Optional[Callable] = None,
) -> NDArray[float]:
    """Numerical Hessian from central finite gradient differences.

    See central differences in
      https://en.wikipedia.org/wiki/Finite_difference_coefficient
    for the different accuracies.
    """
    if callback is None:

        def callback(*args):
            pass

    accuracies = {
        2: ((-0.5, -1), (0.5, 1)),  # 2 calculations
        4: ((1 / 12, -2), (-2 / 3, -1), (2 / 3, 1), (-1 / 12, 2)),  # 4 calculations
    }
    accs_avail = list(accuracies.keys())
    assert acc in accs_avail

    size = coords.size
    fd_hessian = np.zeros((size, size))
    zero_step = np.zeros(size)

    coeffs = accuracies[acc]
    for i, _ in enumerate(coords):
        step = zero_step.copy()
        step[i] = step_size

        def get_grad(factor, displ, j):
            displ_coords = coords + step * displ
            callback(i, j)
            grad = grad_func(displ_coords)
            return factor * grad

        grads = [get_grad(factor, displ, j) for j, (factor, displ) in enumerate(coeffs)]
        fd = np.sum(grads, axis=0) / step_size
        fd_hessian[i] = fd

    # Symmetrize
    fd_hessian = (fd_hessian + fd_hessian.T) / 2
    return fd_hessian


def rot_quaternion(coords3d, ref_coords3d):
    # Translate to origin by removing centroid
    c3d = coords3d - coords3d.mean(axis=0)
    ref_c3d = ref_coords3d - ref_coords3d.mean(axis=0)

    # Setup correlation matrix
    R = c3d.T.dot(ref_c3d)

    # Setup F matrix, Eq. (6) in [1]
    F = np.zeros((4, 4))
    R11, R12, R13, R21, R22, R23, R31, R32, R33 = R.flatten()
    # Fill only upper triangular part.
    F[0, 0] = R11 + R22 + R33
    F[0, 1] = R23 - R32
    F[0, 2] = R31 - R13
    F[0, 3] = R12 - R21
    #
    F[1, 1] = R11 - R22 - R33
    F[1, 2] = R12 + R21
    F[1, 3] = R13 + R31
    #
    F[2, 2] = -R11 + R22 - R33
    F[2, 3] = R23 + R32
    #
    F[3, 3] = -R11 - R22 + R33

    # Eigenvalues, eigenvectors of upper triangular part.
    w, v_ = np.linalg.eigh(F, UPLO="U")

    # Quaternion corresponds to biggest (last) eigenvalue.
    # np.linalg.eigh already returns sorted eigenvalues.
    return w, v_, c3d, ref_c3d


def quaternion_to_rot_mat(q):
    q0, q1, q2, q3 = q
    q_ = q0**2 - (q1**2 + q2**2 + q3**2)
    R = np.zeros((3, 3))
    R[0, 0] = q_ + 2 * q1**2
    R[0, 1] = 2 * (q1 * q2 - q0 * q3)
    R[0, 2] = 2 * (q1 * q3 - q0 * q2)
    R[1, 0] = 2 * (q1 * q2 + q0 * q3)
    R[1, 1] = q_ + 2 * q2**2
    R[1, 2] = 2 * (q2 * q3 - q0 * q1)
    R[2, 0] = 2 * (q1 * q3 - q0 * q2)
    R[2, 1] = 2 * (q2 * q3 + q0 * q1)
    R[2, 2] = q_ + 2 * q3**2
    return R


def rmsd_grad(
    coords3d: NDArray[float], ref_coords3d: NDArray[float], offset: float = 1e-9
) -> Tuple[float, NDArray[float]]:
    """RMSD and gradient between two sets of coordinates from quaternions.

    The gradient is given w.r.t. the coordinates of 'coords3d'.

    Python adaption of
        ls_rmsd.f90
    from the xtb repository of the Grimme group, which in turn implements
        [1] https://doi.org/10.1002/jcc.20110
    .

    Parameters
    ----------
    coords3d
        Coordinate array of shape (N, 3) with N denoting the number of atoms.
    ref_coords3d
        Reference coordinates.
    offset
        Small floating-point number that is added to the RMSD, to avoid division
        by zero.

    Returns
    -------
    rmsd
        RMSD value.
    rmsd_grad
        Gradient of the RMSD value w.r.t. to 'coords3d'.
    """
    assert coords3d.shape == ref_coords3d.shape

    w, v_, c3d, ref_c3d = rot_quaternion(coords3d, ref_coords3d)
    quat = v_[:, -1]
    eigval = w[-1]

    atom_num = coords3d.shape[0]
    x_norm = np.linalg.norm(c3d) ** 2
    y_norm = np.linalg.norm(ref_c3d) ** 2
    rmsd = np.sqrt(max(0.0, ((x_norm + y_norm) - 2.0 * eigval)) / (atom_num)) + offset

    # scipy expects the quaternion
    #     a + b*i + c*j + d*k
    # as (i, j, k, a), that is the scalar component must come last.
    # Currently we have (a, i, j, k), so we have to rearrange before passing the
    # quaternion to scipy.
    rot = Rotation.from_quat((*quat[1:], quat[0]))
    U = rot.as_matrix()

    grad = (c3d - ref_c3d @ U) / (rmsd * atom_num)
    return rmsd, grad


def pivoted_cholesky(A: NDArray, tol: float = -1.0):
    """Cholesky factorization a real symmetric positive semidefinite matrix.
    Cholesky factorization is carried out with full pivoting.

    Adapated from PySCF.

    P.T * A * P = L * L.T

    Parameters
    ----------
    A
        Matrix to be factorized.
    tol
        User defined tolerance, as outlined in the LAPACK docs.

    Returns
    -------
    L
        Lower or upper triangular matrix.
    piv
        Pivot vectors, starting at 0.
    rank
        Rank of the factoirzed matrix.
    """

    N = A.shape[0]
    L, piv, rank, info = dpstrf(A, tol=tol, lower=True)
    piv -= 1  # LAPACK returns 1-based indices
    assert info >= 0
    L[np.triu_indices(N, k=1)] = 0
    L[:, rank:] = 0
    return L, piv, rank


def pivoted_cholesky2(A: NDArray[float], tol: Optional[float] = None, m_max: int = 0):
    """https://doi.org/10.1016/j.apnum.2011.10.001"""

    R = np.zeros_like(A)
    m = 0
    n = len(A)
    # Decompose to max rank.
    if tol is None:
        tol = 0.0
    assert tol >= 0.0, f"{tol=} must be >= 0.0!"
    if m_max == 0:
        m_max = n
    assert 0 <= m_max <= n, f"{m_max=} must fullfil (0 <= m_max <= {n=})!"
    d = np.diag(A).copy()  # Diagonal
    error = np.linalg.norm(A, ord="nuc")
    piv = np.arange(n, dtype=int)
    while True:
        # Stop when error is sufficiently small or we are at max rank.
        # while (error > tol) or (m_max and m < m_max):
        if (error <= tol) or (m_max and (m == m_max)):
            break
        i = m + d[piv[m:]].argmax()
        piv[m], piv[i] = piv[i], piv[m]
        R[m, piv[m]] = np.sqrt(d[piv[m]])
        for i in range(m + 1, n):
            j = np.arange(m)
            sum_ = (R[j, piv[m]] * R[j, piv[i]]).sum()
            R[m, piv[i]] = (A[piv[m], piv[i]] - sum_) / R[m, piv[m]]
            d[piv[i]] -= R[m, piv[i]] ** 2
        error = d[m + 1 :].sum()
        m += 1
    # R.T @ R == A
    return R, piv, m


def matrix_power(mat, p, thresh=1e-12, strict=True):
    w, v = np.linalg.eigh(mat)
    assert (not strict) or (
        w > 0.0
    ).all(), "matrix_power must be called with a (semi)-positive-definite matrix!"
    mask = np.abs(w) > thresh
    w_pow = w[mask] ** p
    v_mask = v[:, mask]
    return v_mask @ np.diag(w_pow) @ v_mask.T


def sym_mat_from_triu(arr, data):
    nrows, ncols = arr.shape
    assert nrows == ncols
    triu = np.triu_indices(nrows)
    tril1 = np.tril_indices(nrows, k=-1)
    arr[triu] = data
    arr[tril1] = arr.T[tril1]


def sym_mat_from_tril(arr, data):
    nrows, ncols = arr.shape
    assert nrows == ncols
    tril = np.tril_indices(nrows)
    arr[tril] = data
    triu1 = np.triu_indices(nrows, k=1)
    arr[triu1] = arr.T[triu1]


def multi_component_sym_mat(arr, dim):
    target_shape = (dim, dim)
    shape = arr.shape
    sym = np.zeros((*target_shape, *shape[1:]))
    triu = np.triu_indices(dim)
    triu1 = np.triu_indices(dim, k=1)
    tril1 = np.tril_indices(dim, k=-1)
    sym[triu] = arr
    sym[tril1] = sym[triu1]
    return sym.reshape(*target_shape, *shape[1:])



"""Smoke regressions for previously low-coverage utility subcommands."""

from __future__ import annotations

from pathlib import Path

from click.testing import CliRunner

import pdb2reaction.energy_diagram as energy_diagram
from pdb2reaction.cli import cli as root_cli


def _write_text(path: Path, text: str) -> Path:
    path.write_text(text, encoding="utf-8")
    return path


def test_add_elem_info_smoke(tmp_path: Path) -> None:
    in_pdb = _write_text(
        tmp_path / "input_no_elem.pdb",
        (
            "ATOM      1  CA  ALA A   1       0.000   0.000   0.000  1.00 10.00\n"
            "END\n"
        ),
    )
    out_pdb = tmp_path / "output_add_elem.pdb"

    runner = CliRunner()
    result = runner.invoke(
        root_cli,
        ["add-elem-info", "-i", str(in_pdb), "-o", str(out_pdb)],
        catch_exceptions=False,
    )

    assert result.exit_code == 0, result.output
    assert out_pdb.exists()

    atom_line = next(line for line in out_pdb.read_text(encoding="utf-8").splitlines() if line.startswith("ATOM"))
    assert len(atom_line) >= 78
    assert atom_line[76:78].strip() == "C"


def test_fix_altloc_smoke(tmp_path: Path) -> None:
    in_pdb = _write_text(
        tmp_path / "altloc_input.pdb",
        (
            "ATOM      1  CA AALA A   1       0.000   0.000   0.000  0.60 10.00           C\n"
            "ATOM      2  CA BALA A   1       0.100   0.000   0.000  0.40 10.00           C\n"
            "END\n"
        ),
    )
    out_pdb = tmp_path / "altloc_clean.pdb"

    runner = CliRunner()
    result = runner.invoke(
        root_cli,
        ["fix-altloc", "-i", str(in_pdb), "-o", str(out_pdb)],
        catch_exceptions=False,
    )

    assert result.exit_code == 0, result.output
    assert out_pdb.exists()

    atom_lines = [line for line in out_pdb.read_text(encoding="utf-8").splitlines() if line.startswith("ATOM")]
    assert len(atom_lines) == 1
    assert atom_lines[0][16] == " "


def test_trj2fig_csv_smoke(tmp_path: Path) -> None:
    trj = _write_text(
        tmp_path / "traj.xyz",
        (
            "1\n"
            "0.000000\n"
            "H 0.0 0.0 0.0\n"
            "1\n"
            "0.500000\n"
            "H 0.0 0.0 0.1\n"
        ),
    )
    out_csv = tmp_path / "energy.csv"

    runner = CliRunner()
    result = runner.invoke(
        root_cli,
        ["trj2fig", "-i", str(trj), "-o", str(out_csv), "--unit", "kcal", "-r", "init"],
        catch_exceptions=False,
    )

    assert result.exit_code == 0, result.output
    assert out_csv.exists()

    lines = out_csv.read_text(encoding="utf-8").splitlines()
    assert lines[0] == "frame,energy_hartree,delta_kcal"
    assert len(lines) == 3


def test_energy_diagram_smoke(tmp_path: Path, monkeypatch) -> None:
    out_png = tmp_path / "energy_diagram.png"

    class _DummyFigure:
        def write_image(self, out_path: str, scale: int = 2) -> None:
            _ = scale
            Path(out_path).write_text("dummy-image", encoding="utf-8")

    monkeypatch.setattr(
        energy_diagram,
        "build_energy_diagram",
        lambda **_kwargs: _DummyFigure(),
    )

    runner = CliRunner()
    result = runner.invoke(
        root_cli,
        [
            "energy-diagram",
            "-i",
            "0",
            "-i",
            "1.2",
            "-o",
            str(out_png),
            "--label-x",
            "R",
            "--label-x",
            "TS",
        ],
        catch_exceptions=False,
    )

    assert result.exit_code == 0, result.output
    assert out_png.exists()
    assert out_png.read_text(encoding="utf-8") == "dummy-image"



# [1] https://aip.scitation.org/doi/pdf/10.1063/1.1498468
#     Bour, 2002

import numpy as np

from pysisyphus.optimizers.HessianOptimizer import HessianOptimizer
from pysisyphus.helpers_pure import eigval_to_wavenumber 


class NCOptimizer(HessianOptimizer):

    def __init__(self, geometry, *args, freeze_modes=None, **kwargs):
        super().__init__(geometry, **kwargs)

        assert not self.is_cos and self.geometry.coord_type == "cart", \
            "NCOptimizer can't be used with ChainOfStates-methods and " \
            "coordinate systems beside cartesians ('coord_type: cart')."

        self.freeze_modes = freeze_modes
        
        self.sqrt_m = np.sqrt(self.geometry.masses_rep)

    def optimize(self):
        grad = self.geometry.gradient
        self.forces.append(-grad.copy())
        self.energies.append(self.geometry.energy)

        if self.cur_cycle > 0:
            self.update_trust_radius()
            self.update_hessian()

        # eigvals_org, eigvecs_org = np.linalg.eigh(self.H)
        # grad_trans = eigvals_org.T.dot(grad)

        # Mass-weighted hessian
        H_mw = self.geometry.mass_weigh_hessian(self.H)
        # Project out translation and rotation
        if self.geometry.coords.size > 3:
            H_mw = self.geometry.eckart_projection(H_mw)
        eigvals, eigvecs = np.linalg.eigh(H_mw)
        # Drop translational/rotational modes as they will have
        # small (zero) eigenvalues. Zero the corresponding gradient entries.
        eigvals, eigvecs, small_inds = self.filter_small_eigvals(eigvals, eigvecs, mask=True)

        wavenumbers = eigval_to_wavenumber(eigvals)

        freeze_inds = np.zeros_like(wavenumbers, dtype=bool)
        if self.freeze_modes:
            freeze_inds = wavenumbers < self.freeze_modes
            eigvals = eigvals[~freeze_inds]
            eigvecs = eigvecs[:,~freeze_inds]
            self.log(f"{np.sum(freeze_inds)} normal modes will be frozen.")

        frozen_str = ["(frozen)" if frozen else "" for frozen in freeze_inds]
        wavenumber_str = "\n".join([
                            f"\t{i:> 3d}: {wn:> 8.2f} cm⁻¹ {frz}"
                            for i, (wn, frz) in enumerate(zip(wavenumbers, frozen_str))
        ])
        self.log("Frequencies:\n" + wavenumber_str)

        # Transform gradient to eigensystem of Hessian. We also have to use
        # the mass-weighted gradient.
        grad_q = eigvecs.T.dot(self.geometry.mw_gradient)

        mw_H_aug = self.get_augmented_hessian(eigvals, grad_q)
        # Discard eigenvector for now
        mw_step, eigval, nu, _ = self.solve_rfo(mw_H_aug, "min")
        # Transform back to original basis. Right now the step is still in
        # mass-weighted coordinates.
        mw_step = eigvecs @ mw_step
        # Un-massweigh step
        step = mw_step / self.sqrt_m

        step_norm = np.linalg.norm(step)

        if step_norm > self.trust_radius:
            step = step / step_norm * self.trust_radius

        if self.freeze_modes:
            # With frozen modes we only want to consider the gradient contributions
            # from non-frozen modes. So here we transform back the gradient and
            # un-weigh it.
            grad = eigvecs.dot(grad_q) * self.sqrt_m
            self.modified_forces.append(-grad)

        quadratic_prediction = step @ grad + 0.5 * step @ self.H @ step
        rfo_prediction = quadratic_prediction / (1 + step @ step)
        self.predicted_energy_changes.append(rfo_prediction)

        return step
